# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import (
    FileFormatConfig,
    IOConfig,
    StorageConfig,
    WarcSourceConfig,
)
from daft.dataframe import DataFrame
from daft.datatype import DataType, TimeUnit
from daft.io.common import get_tabular_files_scan


@PublicAPI
def read_warc(
    path: Union[str, list[str]],
    io_config: Optional[IOConfig] = None,
    file_path_column: Optional[str] = None,
    _multithreaded_io: Optional[bool] = None,
) -> DataFrame:
    """Creates a DataFrame from WARC or gzipped WARC file(s). This is an experimental feature and the API may change in the future.

    Args:
        path (Union[str, List[str]]): Path to WARC file (allows for wildcards)
        io_config (Optional[IOConfig]): Config to be used with the native downloader
        file_path_column (Optional[str]): Include the source path(s) as a column with this name. Defaults to None.
        _multithreaded_io (Optional[bool]): Whether to use multithreading for IO threads. Setting this to False can be helpful in reducing
            the amount of system resources (number of connections and thread contention) when running in the Ray runner.
            Defaults to None, which will let Daft decide based on the runner it is currently using.

    Returns:
        DataFrame: parsed DataFrame with mandatory metadata columns ("WARC-Record-ID", "WARC-Type", "WARC-Date", "Content-Length"), one optional
            metadata column ("WARC-Identified-Payload-Type"), one column "warc_content" with the raw byte content of the WARC record,
            and one column "warc_headers" with the remaining headers of the WARC record stored as a JSON string.

    Examples:
        >>> df = daft.read_warc("/path/to/file.warc")
        >>> df = daft.read_warc("/path/to/directory")
        >>> df = daft.read_warc("/path/to/files-*.warc")
        >>> df = daft.read_warc("s3://path/to/files-*.warc")
        >>> df = daft.read_warc("gs://path/to/files-*.warc")

    """
    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config

    if isinstance(path, list) and len(path) == 0:
        raise ValueError("Cannot read DataFrame from from empty list of Warc filepaths")

    # If running on Ray, we want to limit the amount of concurrency and requests being made.
    # This is because each Ray worker process receives its own pool of thread workers and connections.
    multithreaded_io = (
        (context.get_context().get_or_create_runner().name != "ray") if _multithreaded_io is None else _multithreaded_io
    )
    storage_config = StorageConfig(multithreaded_io, io_config)

    schema = {
        "WARC-Record-ID": DataType.string(),
        "WARC-Type": DataType.string(),
        "WARC-Date": DataType.timestamp(TimeUnit.ns(), timezone="Etc/UTC"),
        "Content-Length": DataType.int64(),
        "WARC-Identified-Payload-Type": DataType.string(),
        "warc_content": DataType.binary(),
        "warc_headers": DataType.string(),
    }

    warc_config = WarcSourceConfig()
    file_format_config = FileFormatConfig.from_warc_config(warc_config)

    builder = get_tabular_files_scan(
        path=path,
        infer_schema=False,
        schema=schema,
        file_format_config=file_format_config,
        storage_config=storage_config,
        file_path_column=file_path_column,
        hive_partitioning=False,
    )
    return DataFrame(builder)



from __future__ import annotations

import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.functions import monotonically_increasing_id
from daft.io._generator import read_generator
from daft.recordbatch.recordbatch import RecordBatch
from tests.conftest import get_tests_daft_runner_name


def test_monotonically_increasing_id_single_partition(make_df) -> None:
    data = {"a": [1, 2, 3, 4, 5]}
    df = make_df(data)._add_monotonically_increasing_id().collect()

    assert len(df) == 5
    assert set(df.column_names) == {"id", "a"}
    assert df.schema()["id"].dtype == DataType.uint64()
    assert df.to_pydict() == {"id": [0, 1, 2, 3, 4], "a": [1, 2, 3, 4, 5]}

    # Test new function matches old behavior
    df2 = make_df(data).with_column("id", monotonically_increasing_id()).collect()
    assert df.to_pydict() == df2.to_pydict()


def test_monotonically_increasing_id_empty_table(make_df) -> None:
    data = {"a": []}
    df = make_df(data)._add_monotonically_increasing_id().collect()

    assert len(df) == 0
    assert set(df.column_names) == {"id", "a"}
    assert df.schema()["id"].dtype == DataType.uint64()
    assert df.to_pydict() == {"id": [], "a": []}

    # Test new function matches old behavior
    df2 = make_df(data).with_column("id", monotonically_increasing_id()).collect()
    assert df.to_pydict() == df2.to_pydict()


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "native",
    reason="Native runner does not support repartitioning",
)
@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_monotonically_increasing_id_multiple_partitions_with_into_partition(make_df, repartition_nparts) -> None:
    ITEMS = [i for i in range(100)]

    data = {"a": ITEMS}
    df = make_df(data).into_partitions(repartition_nparts)._add_monotonically_increasing_id().collect()

    assert len(df) == 100
    assert set(df.column_names) == {"id", "a"}
    assert df.schema()["id"].dtype == DataType.uint64()

    # Test new function matches old behavior
    df2 = make_df(data).into_partitions(repartition_nparts).with_column("id", monotonically_increasing_id()).collect()
    assert df.to_pydict() == df2.to_pydict()

    # we can predict the ids because into_partitions evenly distributes without shuffling the data,
    # and the chosen repartition_nparts is a multiple of the number of items, so each partition will have the same number of items
    items_per_partition = len(ITEMS) // repartition_nparts
    ids = []
    for index, _ in enumerate(ITEMS):
        partition_num = index // items_per_partition
        counter = index % items_per_partition
        ids.append(partition_num << 36 | counter)

    assert df.to_pydict() == {"id": ids, "a": ITEMS}


def test_monotonically_increasing_id_from_generator() -> None:
    ITEMS = list(range(10))
    table = RecordBatch.from_pydict({"a": ITEMS})

    num_tables = 3
    num_generators = 3

    def generator():
        for _ in range(num_tables):
            yield table

    def generators():
        for _ in range(num_generators):
            yield generator

    df = read_generator(generators(), schema=table.schema())._add_monotonically_increasing_id().collect()

    assert len(df) == 90
    assert set(df.column_names) == {"id", "a"}
    assert df.schema()["id"].dtype == DataType.uint64()

    # Test new function matches old behavior
    df2 = read_generator(generators(), schema=table.schema()).with_column("id", monotonically_increasing_id()).collect()
    assert df.to_pydict() == df2.to_pydict()

    if get_tests_daft_runner_name() == "native":
        # On the native runner, there are no partitions, so the ids are just the row numbers.
        assert df.to_pydict() == {"id": list(range(90)), "a": ITEMS * 9}
    else:
        # On the ray / py runner, the ids are generated based on the partition number and the row number within the partition.
        # The partition number is put in the upper 28 bits and the row number is put in the lower 36 bits.
        # There are num_generators partitions, and each partition has num_tables * len(ITEMS) rows.
        ids = [(p << 36) | c for p in range(num_generators) for c in range(num_tables * len(ITEMS))]
        assert df.to_pydict() == {"id": ids, "a": ITEMS * 9}


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_monotonically_increasing_id_multiple_partitions_with_repartition(make_df, repartition_nparts) -> None:
    ITEMS = [i for i in range(100)]

    data = {"a": ITEMS}
    df = make_df(data, repartition=repartition_nparts)._add_monotonically_increasing_id().collect()

    assert len(df) == 100
    assert set(df.column_names) == {"id", "a"}
    assert df.schema()["id"].dtype == DataType.uint64()

    py_dict = df.to_pydict()
    assert set(py_dict["a"]) == set(ITEMS)

    # cannot predict the ids because repartition shuffles the data, so we just check that they are unique
    assert len(set(py_dict["id"])) == 100

    # Test new function matches old behavior
    df2 = make_df(data, repartition=repartition_nparts).with_column("id", monotonically_increasing_id()).collect()

    assert len(df2) == 100
    assert set(df2.column_names) == {"id", "a"}
    assert df2.schema()["id"].dtype == DataType.uint64()

    py_dict2 = df2.to_pydict()
    assert set(py_dict2["a"]) == set(ITEMS)
    assert len(set(py_dict2["id"])) == 100


def test_monotonically_increasing_id_custom_col_name(make_df) -> None:
    data = {"a": [1, 2, 3, 4, 5]}
    df = make_df(data)._add_monotonically_increasing_id("custom_id").collect()

    assert len(df) == 5
    assert set(df.column_names) == {"custom_id", "a"}
    assert df.schema()["custom_id"].dtype == DataType.uint64()
    assert df.to_pydict() == {"custom_id": [0, 1, 2, 3, 4], "a": [1, 2, 3, 4, 5]}

    # Test new function matches old behavior with custom column name
    df2 = make_df(data).with_column("custom_id", monotonically_increasing_id()).collect()
    assert df.to_pydict() == df2.to_pydict()


def test_monotonic_id_with_complex_projections(make_df) -> None:
    """Test monotonically_increasing_id with complex projections to verify optimization rule."""
    data = {"a": [1, 2, 3], "b": [4, 5, 6]}

    # Create DataFrame with multiple column operations
    df1 = (
        make_df(data)
        .with_columns(
            {"id": monotonically_increasing_id(), "a_plus_b": col("a") + col("b"), "b_squared": col("b") * col("b")}
        )
        .collect()
    )

    # Verify schema
    assert set(df1.column_names) == {"id", "a", "b", "a_plus_b", "b_squared"}
    assert df1.schema()["id"].dtype == DataType.uint64()

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        ._add_monotonically_increasing_id()
        .with_columns({"a_plus_b": col("a") + col("b"), "b_squared": col("b") * col("b")})
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_with_filter_operations(make_df) -> None:
    """Test monotonically_increasing_id with filter operations to verify optimization rule."""
    data = {"a": list(range(10))}

    # Create DataFrame with filter before and after
    df1 = (
        make_df(data)
        .filter(col("a") > 3)  # Filter before
        .with_column("id", monotonically_increasing_id())
        .filter(col("id") < 5)  # Filter after
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        .filter(col("a") > 3)  # Filter before
        ._add_monotonically_increasing_id()
        .filter(col("id") < 5)  # Filter after
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_with_mixed_operations(make_df) -> None:
    """Test monotonically_increasing_id with a mix of operations to verify optimization rule."""
    data = {"key": ["a", "b", "a", "b", "c"], "value": [1, 2, 3, 4, 5]}

    # Create DataFrame with mixed operations
    df1 = (
        make_df(data)
        .into_partitions(2)
        .with_column("id", monotonically_increasing_id())
        .groupby("key")
        .agg(col("value").sum().alias("sum"), col("id").min().alias("first_id"))
        .sort("key")
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        .into_partitions(2)
        ._add_monotonically_increasing_id()
        .groupby("key")
        .agg(col("value").sum().alias("sum"), col("id").min().alias("first_id"))
        .sort("key")
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_chained_operations(make_df) -> None:
    """Test multiple monotonically_increasing_id operations in a chain."""
    data = {"a": [1, 2, 3]}

    # Create DataFrame with multiple ID columns
    df1 = (
        make_df(data)
        .with_column("id1", monotonically_increasing_id())
        .with_column("id2", monotonically_increasing_id())
        .collect()
    )

    # Create same DataFrame using old method
    df2 = make_df(data)._add_monotonically_increasing_id("id1")._add_monotonically_increasing_id("id2").collect()

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_with_join(make_df) -> None:
    """Test monotonically_increasing_id with join operations."""
    left_data = {"key": ["a", "b", "c"], "value": [1, 2, 3]}
    right_data = {"key": ["b", "c", "d"], "other": [4, 5, 6]}

    # Create DataFrames with ID before join
    df1 = (
        make_df(left_data)
        .with_column("id", monotonically_increasing_id())
        .join(make_df(right_data).with_column("id", monotonically_increasing_id()), on="key", how="outer")
        .sort("key")
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(left_data)
        ._add_monotonically_increasing_id()
        .join(make_df(right_data)._add_monotonically_increasing_id(), on="key", how="outer")
        .sort("key")
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_with_complex_expressions(make_df) -> None:
    """Test monotonically_increasing_id with complex expressions to verify optimization rule handles nested expressions."""
    data = {"a": [1, 2, 3]}

    # Create DataFrame with monotonically_increasing_id in complex expressions
    df1 = (
        make_df(data)
        .with_columns(
            {
                "id_plus_one": monotonically_increasing_id() + 1,
                "id_times_a": monotonically_increasing_id() * col("a"),
                "id_conditional": (monotonically_increasing_id() > 1).if_else(col("a") * 2, col("a")),
            }
        )
        .collect()
    )

    # Verify schema
    assert set(df1.column_names) == {"a", "id_plus_one", "id_times_a", "id_conditional"}

    # Create same DataFrame using old method with manual calculations
    df2 = (
        make_df(data)
        ._add_monotonically_increasing_id()
        .with_columns(
            {
                "id_plus_one": col("id") + 1,
                "id_times_a": col("id") * col("a"),
                "id_conditional": (col("id") > 1).if_else(col("a") * 2, col("a")),
            }
        )
        .select("a", "id_plus_one", "id_times_a", "id_conditional")  # Pass column names as separate arguments
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_multiple_calls_same_expression(make_df) -> None:
    """Test using multiple monotonically_increasing_id calls in the same expression."""
    data = {"a": [1, 2, 3]}

    # Create DataFrame with multiple monotonically_increasing_id calls in same expression
    df1 = (
        make_df(data)
        .with_columns(
            {
                "id_sum": monotonically_increasing_id() + monotonically_increasing_id(),
                "id_product": monotonically_increasing_id() * monotonically_increasing_id(),
                "id_compare": monotonically_increasing_id() > monotonically_increasing_id(),
            }
        )
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        ._add_monotonically_increasing_id("id1")
        ._add_monotonically_increasing_id("id2")
        .with_columns(
            {
                "id_sum": col("id1") + col("id2"),
                "id_product": col("id1") * col("id2"),
                "id_compare": col("id1") > col("id2"),
            }
        )
        .select("a", "id_sum", "id_product", "id_compare")
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_deeply_nested(make_df) -> None:
    """Test deeply nested expressions containing monotonically_increasing_id."""
    data = {"a": [1, 2, 3], "b": [4, 5, 6]}

    # Create DataFrame with deeply nested expressions
    df1 = (
        make_df(data)
        .with_columns(
            {
                "nested1": (monotonically_increasing_id() + col("a")) * (monotonically_increasing_id() + col("b")),
                "nested2": ((monotonically_increasing_id() > 0) & (col("a") > 2)).if_else(
                    monotonically_increasing_id() * 2, monotonically_increasing_id() + 1
                ),
            }
        )
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        ._add_monotonically_increasing_id("id1")
        ._add_monotonically_increasing_id("id2")
        ._add_monotonically_increasing_id("id3")
        .with_columns(
            {
                "nested1": (col("id1") + col("a")) * (col("id2") + col("b")),
                "nested2": ((col("id1") > 0) & (col("a") > 2)).if_else(col("id2") * 2, col("id3") + 1),
            }
        )
        .select("a", "b", "nested1", "nested2")
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_complex_conditionals(make_df) -> None:
    """Test complex conditional expressions with monotonically_increasing_id."""
    data = {"value": [1, 2, 3, 4, 5]}

    # Create DataFrame with complex conditional logic
    df1 = (
        make_df(data)
        .with_columns(
            {
                "case1": (monotonically_increasing_id() % 2 == 0).if_else(col("value") * 2, col("value")),
                "case2": (
                    (monotonically_increasing_id() < 2).if_else(
                        col("value"), (monotonically_increasing_id() < 4).if_else(col("value") * 2, col("value") * 3)
                    )
                ),
            }
        )
        .collect()
    )

    # Create same DataFrame using old method
    df2 = (
        make_df(data)
        ._add_monotonically_increasing_id()
        .with_columns(
            {
                "case1": (col("id") % 2 == 0).if_else(col("value") * 2, col("value")),
                "case2": (
                    (col("id") < 2).if_else(col("value"), (col("id") < 4).if_else(col("value") * 2, col("value") * 3))
                ),
            }
        )
        .select("value", "case1", "case2")
        .collect()
    )

    # Verify both paths produce identical results
    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_in_filter(make_df) -> None:
    """Test using monotonically_increasing_id in filter conditions."""
    data = {"a": [1, 2, 3, 4, 5]}

    # Use _add_monotonically_increasing_id
    df1 = make_df(data)._add_monotonically_increasing_id().filter(col("id") > 2).collect()

    # Filter using monotonically_increasing_id directly in filter should raise an error
    with pytest.raises(Exception, match="monotonically_increasing_id\\(\\) is only allowed in projections"):
        make_df(data).filter(monotonically_increasing_id() > 2).collect()

    # This approach works - add column first, then filter
    df2 = make_df(data).with_column("id", monotonically_increasing_id()).filter(col("id") > 2).collect()

    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_in_aggregation(make_df) -> None:
    """Test using monotonically_increasing_id in aggregation expressions."""
    data = {"key": ["a", "b", "a", "b", "c"], "value": [1, 2, 3, 4, 5]}

    # Use _add_monotonically_increasing_id
    df1 = (
        make_df(data)
        ._add_monotonically_increasing_id()
        .groupby("key")
        .agg((col("id") + col("value")).sum().alias("sum_id_plus_value"), col("id").max().alias("max_id"))
        .sort("key")
        .collect()
    )

    # Using monotonically_increasing_id directly in aggregation should raise an error
    with pytest.raises(Exception, match="monotonically_increasing_id\\(\\) is only allowed in projections"):
        (
            make_df(data)
            .groupby("key")
            .agg(
                (monotonically_increasing_id() + col("value")).sum().alias("sum_id_plus_value"),
                monotonically_increasing_id().max().alias("max_id"),
            )
            .sort("key")
            .collect()
        )

    # This approach works - add column first, then aggregate
    df2 = (
        make_df(data)
        .with_column("id", monotonically_increasing_id())
        .groupby("key")
        .agg((col("id") + col("value")).sum().alias("sum_id_plus_value"), col("id").max().alias("max_id"))
        .sort("key")
        .collect()
    )

    assert df1.to_pydict() == df2.to_pydict()


def test_monotonic_id_in_join_condition(make_df) -> None:
    """Test using monotonically_increasing_id in join conditions."""
    left_data = {"key": ["a", "b", "c"], "value": [1, 2, 3]}
    right_data = {"key": ["b", "c", "d"], "other": [4, 5, 6]}

    # Use _add_monotonically_increasing_id
    df1 = (
        make_df(left_data)
        ._add_monotonically_increasing_id()
        .join(make_df(right_data)._add_monotonically_increasing_id(), on="id", how="inner")
        .collect()
    )

    # Using monotonically_increasing_id directly in join condition should raise an error
    with pytest.raises(Exception, match="monotonically_increasing_id\\(\\) is only allowed in projections"):
        (
            make_df(left_data)
            .join(make_df(right_data), on=monotonically_increasing_id() == monotonically_increasing_id(), how="inner")
            .collect()
        )

    # This approach works - add column first, then join
    df2 = (
        make_df(left_data)
        .with_column("id", monotonically_increasing_id())
        .join(make_df(right_data).with_column("id", monotonically_increasing_id()), on="id", how="inner")
        .collect()
    )

    assert df1.to_pydict() == df2.to_pydict()


def test_monotonically_increasing_id_with_cast(make_df) -> None:
    """Test that casting monotonically_increasing_id directly works correctly.

    This tests addresses a bug where calling .cast() directly on monotonically_increasing_id()
    would result in a PanicException with 'called `Option::unwrap()` on a `None` value'.
    """
    data = {"foo": list(range(10))}

    df = make_df(data).with_column("id", monotonically_increasing_id().cast(DataType.string())).collect()

    assert len(df) == 10
    assert df.schema()["id"].dtype == DataType.string()

    df2 = (
        make_df(data)
        .with_column("id", monotonically_increasing_id())
        .with_column("id_str", col("id").cast(DataType.string()))
        .collect()
    )

    assert df.to_pydict()["id"] == df2.to_pydict()["id_str"]



from __future__ import annotations

import pandas as pd
import pytest

import daft
from daft import col

TESTS = [
    [[], 0],
    [[None] * 10, 0],
    [[1], 1],
    [["hello"], 1],
    [[1, 2, 3, 4, 3, 2, 1, None, None, None], 4],
    [[10] * 10, 1],
    [["Hello", "hello", "he", "hello"], 3],
    [[True, False, True, False], 2],
]


def make_df(list: list):
    return daft.from_pydict({"a": list})


def make_asymmetric_df(list) -> daft.DataFrame:
    df = daft.from_pydict({"a": list})

    if list:
        # If I just did `daft.from_pydict({ "a": [] })`, I would get a schema of type `Null`.
        # So instead, I create it with a value and then filter against it (thus giving me an empty df).
        first = list[0]
        empty_df = daft.from_pydict({"a": [first]}).where(col("a") != first).collect()
    else:
        # The list is empty so it doesn't matter what the schema is.
        empty_df = daft.from_pydict({"a": []})

    return df.concat(empty_df).into_partitions(2).collect()


def assert_equal(df: daft.DataFrame, expected):
    pd.testing.assert_series_equal(
        pd.Series(df.to_pydict()["a"]),
        pd.Series(expected),
        check_exact=True,
    )


# Purpose:
# Test regular columns with varying sizes of partitions.
# We want to test cases in which the "singular partition" optimization is hit and cases where it's not.
@pytest.mark.parametrize("data_and_expected", TESTS)
@pytest.mark.parametrize("partition_size", [None, 2, 3])
def test_approx_count_distinct(data_and_expected, partition_size, with_morsel_size):
    data, expected = data_and_expected
    df = make_df(data)
    if partition_size:
        df = df.into_partitions(partition_size).collect()
    df = df.agg(col("a").approx_count_distinct())
    assert_equal(df, expected)


# Purpose:
# Test creating data with empty partitions against the HLL algo.
# We want to test if empty partitions mess with the final global result.
@pytest.mark.parametrize("data_and_expected", TESTS)
def test_approx_count_distinct_on_dfs_with_empty_partitions(data_and_expected, with_morsel_size):
    data, expected = data_and_expected
    df = make_asymmetric_df(data)
    df = df.agg(col("a").approx_count_distinct())
    assert_equal(df, expected)


# Purpose:
# Test the adding `NULL`s values to the existing data.
# We want to test how HLL handles null values.
#
# We should always test `NULL` values as the "absence" of values.
# Therefore, the existence of a `NULL` should never affect the approx_count_distinct value that is returned (even if it's in a column of type `NULL`).
@pytest.mark.parametrize("data_and_expected", TESTS)
def test_approx_count_distinct_on_null_values(data_and_expected, with_morsel_size):
    data, expected = data_and_expected
    data = data + [None] * 10
    df = make_df(data)
    df = df.agg(col("a").approx_count_distinct())
    assert_equal(df, expected)



from __future__ import annotations

import io
import os
from collections.abc import Iterator

import pytest
import requests

from tests.io.mock_aws_server import start_service, stop_process

############################
### AWS-specific fixtures ###
############################


@pytest.fixture(scope="session")
def aws_credentials() -> dict[str, str]:
    return {
        "AWS_ACCESS_KEY_ID": "testing",
        "AWS_SECRET_ACCESS_KEY": "testing",
        "AWS_SESSION_TOKEN": "testing",
    }


@pytest.fixture(scope="session")
def aws_server_ip() -> str:
    return "127.0.0.1"


@pytest.fixture(scope="session")
def aws_server_port() -> int:
    return 5000


@pytest.fixture(scope="session")
def aws_log_file(tmp_path_factory: pytest.TempPathFactory) -> Iterator[io.IOBase]:
    # NOTE(Clark): We have to use a log file for the mock AWS server's stdout/sterr.
    # - If we use None, then the server output will spam stdout.
    # - If we use PIPE, then the server will deadlock if the (relatively small) buffer fills, and the server is pretty
    #   noisy.
    # - If we use DEVNULL, all log output is lost.
    # With a tmp_path log file, we can prevent spam and deadlocks while also providing an avenue for debuggability, via
    # changing this fixture to something persistent, or dumping the file to stdout before closing the file, etc.
    tmp_path = tmp_path_factory.mktemp("aws_logging")
    with open(tmp_path / "aws_log.txt", "w") as f:
        yield f


@pytest.fixture(scope="session")
def aws_server(aws_server_ip: str, aws_server_port: int, aws_log_file: io.IOBase) -> Iterator[str]:
    # NOTE(Clark): The background-threaded moto server tends to lock up under concurrent access, so we run a background
    # moto_server process.
    aws_server_url = f"http://{aws_server_ip}:{aws_server_port}"
    old_env = os.environ.copy()
    # Set required AWS environment variables before starting server.
    # Required to opt out of concurrent writing, since we don't provide a LockClient.
    os.environ["AWS_S3_ALLOW_UNSAFE_RENAME"] = "true"
    try:
        # Start moto server.
        process = start_service(aws_server_ip, aws_server_port, aws_log_file)
        yield aws_server_url
    finally:
        # Shutdown moto server.
        stop_process(process)
        # Restore old set of environment variables.
        os.environ = old_env


@pytest.fixture(scope="function")
def reset_aws(aws_server) -> Iterator[None]:
    # Clears local AWS server of all state (e.g. S3 buckets and objects).
    yield
    requests.post(f"{aws_server}/moto-api/reset")



from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

if TYPE_CHECKING:
    from daft.daft import AzureConfig, GCSConfig, IOConfig, S3Config


def io_config_to_storage_options(io_config: IOConfig, table_uri: str) -> dict[str, str] | None:
    """Converts the Daft IOConfig to a storage options dict that the object_store crate understands.

    The object_store crate is used by many Rust-backed Python libraries such as
    delta-rs and lance.

    This function takes as input the table_uri, which it uses to determine the backend to be used.
    """
    scheme = urlparse(table_uri).scheme
    if scheme == "s3" or scheme == "s3a":
        return _s3_config_to_storage_options(io_config.s3)
    elif scheme == "gcs" or scheme == "gs":
        return _gcs_config_to_storage_options(io_config.gcs)
    elif scheme == "az" or scheme == "abfs" or scheme == "abfss":
        return _azure_config_to_storage_options(io_config.azure)
    else:
        return None


def _s3_config_to_storage_options(s3_config: S3Config) -> dict[str, str]:
    storage_options: dict[str, str] = {}
    if s3_config.region_name is not None:
        storage_options["region"] = s3_config.region_name
    if s3_config.endpoint_url is not None:
        storage_options["endpoint_url"] = s3_config.endpoint_url
    if s3_config.key_id is not None:
        storage_options["access_key_id"] = s3_config.key_id
    if s3_config.session_token is not None:
        storage_options["session_token"] = s3_config.session_token
    if s3_config.access_key is not None:
        storage_options["secret_access_key"] = s3_config.access_key
    if s3_config.use_ssl is not None:
        storage_options["allow_http"] = "false" if s3_config.use_ssl else "true"
    if s3_config.verify_ssl is not None:
        storage_options["allow_invalid_certificates"] = "false" if s3_config.verify_ssl else "true"
    if s3_config.connect_timeout_ms is not None:
        storage_options["connect_timeout"] = str(s3_config.connect_timeout_ms) + "ms"
    if s3_config.anonymous:
        storage_options["skip_signature"] = "true"
    if s3_config.force_virtual_addressing:
        storage_options["virtual_hosted_style_request"] = "true"
    return storage_options


def _azure_config_to_storage_options(azure_config: AzureConfig) -> dict[str, str]:
    storage_options = {}
    if azure_config.storage_account is not None:
        storage_options["account_name"] = azure_config.storage_account
    if azure_config.access_key is not None:
        storage_options["access_key"] = azure_config.access_key
    if azure_config.sas_token is not None:
        storage_options["sas_token"] = azure_config.sas_token
    if azure_config.bearer_token is not None:
        storage_options["bearer_token"] = azure_config.bearer_token
    if azure_config.tenant_id is not None:
        storage_options["tenant_id"] = azure_config.tenant_id
    if azure_config.client_id is not None:
        storage_options["client_id"] = azure_config.client_id
    if azure_config.client_secret is not None:
        storage_options["client_secret"] = azure_config.client_secret
    if azure_config.use_fabric_endpoint is not None:
        storage_options["use_fabric_endpoint"] = "true" if azure_config.use_fabric_endpoint else "false"
    if azure_config.endpoint_url is not None:
        storage_options["endpoint"] = azure_config.endpoint_url
    if azure_config.use_ssl is not None:
        storage_options["allow_http"] = "false" if azure_config.use_ssl else "true"
    return storage_options


def _gcs_config_to_storage_options(gcs_config: GCSConfig) -> dict[str, str]:
    storage_options = {}
    if gcs_config.credentials is not None:
        storage_options["google_application_credentials"] = gcs_config.credentials
    return storage_options



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations
from typing import Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, ScanOperatorHandle
from daft.dataframe import DataFrame
from daft.io.lance.lance_scan import LanceDBScanOperator
from daft.io.object_store_options import io_config_to_storage_options
from daft.logical.builder import LogicalPlanBuilder


@PublicAPI
def read_lance(
    url: str,
    io_config: Optional[IOConfig] = None,
    version: Optional[Union[str, int]] = None,
    asof: Optional[str] = None,
    block_size: Optional[int] = None,
    commit_lock: Optional[object] = None,
    index_cache_size: Optional[int] = None,
    default_scan_options: Optional[dict[str, str]] = None,
    metadata_cache_size_bytes: Optional[int] = None,
) -> DataFrame:
    """Create a DataFrame from a LanceDB table.

    Args:
        url: URL to the LanceDB table (supports remote URLs to object stores such as `s3://` or `gs://`)
        io_config: A custom IOConfig to use when accessing LanceDB data. Defaults to None.

    Returns:
        DataFrame: a DataFrame with the schema converted from the specified LanceDB table

    Note:
        This function requires the use of [LanceDB](https://lancedb.github.io/lancedb/), which is the Python library for the LanceDB project.
        To ensure that this is installed with Daft, you may install: `pip install daft[lance]`

    Examples:
        Read a local LanceDB table:
        >>> df = daft.read_lance("s3://my-lancedb-bucket/data/")
        >>> df.show()

        Read a LanceDB table from a public S3 bucket:
        >>> from daft.io import S3Config
        >>> s3_config = S3Config(region="us-west-2", anonymous=True)
        >>> df = daft.read_lance("s3://daft-public-data/lance/words-test-dataset", io_config=s3_config)
        >>> df.show()

        Read a local LanceDB table and specify a version:
        >>> df = daft.read_lance("s3://my-lancedb-bucket/data/", version=1)
        >>> df.show()
    """
    try:
        import lance
    except ImportError as e:
        raise ImportError(
            "Unable to import the `lance` package, please ensure that Daft is installed with the lance extra dependency: `pip install daft[lance]`"
        ) from e

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_options = io_config_to_storage_options(io_config, url)

    ds = lance.dataset(
        url,
        storage_options=storage_options,
        version=version,
        asof=asof,
        block_size=block_size,
        commit_lock=commit_lock,
        index_cache_size=index_cache_size,
        default_scan_options=default_scan_options,
        metadata_cache_size_bytes=metadata_cache_size_bytes,
    )
    lance_operator = LanceDBScanOperator(ds)

    handle = ScanOperatorHandle.from_python_scan_operator(lance_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)



from __future__ import annotations

import pyarrow as pa
import pytest

from daft import DataType
from daft.series import Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES + ARROW_STRING_TYPES)
def test_series_slice(dtype) -> None:
    data = pa.array([10, 20, 33, None, 50, None])

    s = Series.from_arrow(data.cast(dtype))

    result = s.slice(2, 4)
    assert result.datatype() == s.datatype()
    assert len(result) == 2

    original_data = s.to_pylist()
    expected = original_data[2:4]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("fixed", [False, True])
def test_series_slice_list_array(fixed) -> None:
    dtype = pa.list_(pa.int64(), list_size=2 if fixed else -1)
    data = pa.array([[10, 20], [33, None], [43, 45], None, [50, 52], None], type=dtype)

    s = Series.from_arrow(data)
    result = s.slice(2, 4)
    assert result.datatype() == s.datatype()
    assert len(result) == 2

    if not fixed:
        data = data.cast(pa.large_list(pa.int64()))
    assert result.to_arrow() == data.slice(2, 2)

    original_data = s.to_pylist()
    expected = original_data[2:4]
    assert result.to_pylist() == expected


def test_series_slice_list_array_to_fixed_size_with_nulls() -> None:
    data = [[10, 20], [33, None], [43, 45], None, [50, 52], None]

    s = Series.from_pylist(data)
    result = s.slice(1, 4).cast(DataType.fixed_size_list(DataType.int64(), 2))
    assert result.datatype() == DataType.fixed_size_list(DataType.int64(), 2)
    assert len(result) == 3

    expected = [[33, None], [43, 45], None]
    assert result.to_pylist() == expected


def test_series_slice_list_array_to_fixed_size_without_nulls() -> None:
    data = [[10, 20], [33, 34], [43, 45], [50, 52], [60, 62]]

    s = Series.from_pylist(data)
    result = s.slice(1, 4).cast(DataType.fixed_size_list(DataType.int64(), 2))
    assert result.datatype() == DataType.fixed_size_list(DataType.int64(), 2)
    assert len(result) == 3

    expected = [[33, 34], [43, 45], [50, 52]]
    assert result.to_pylist() == expected


def test_series_slice_struct_array() -> None:
    dtype = pa.struct({"a": pa.int64(), "b": pa.float64()})
    data = pa.array([{"a": 10, "b": 20}, {"a": 33}, {"a": 43, "b": 45}, None, {"a": 50, "b": 52}, None], type=dtype)

    s = Series.from_arrow(data)

    result = s.slice(2, 4)
    assert result.datatype() == s.datatype()
    assert len(result) == 2

    assert result.to_arrow() == data.slice(2, 2)

    original_data = s.to_pylist()
    expected = original_data[2:4]
    assert result.to_pylist() == expected


def test_series_slice_bad_input() -> None:
    data = pa.array([10, 20, 33, None, 50, None])

    s = Series.from_arrow(data)

    with pytest.raises(ValueError, match="slice length can not be negative:"):
        s.slice(3, 2)

    with pytest.raises(ValueError, match="slice start can not be negative"):
        s.slice(-1, 2)

    with pytest.raises(ValueError, match="slice end can not be negative"):
        s.slice(0, -1)



from __future__ import annotations

import os
import tempfile
from collections import defaultdict

import memray
import pytest
from fsspec.implementations.local import LocalFileSystem

import daft
from benchmarking.tpch import data_generation
from tests.assets import TPCH_DBGEN_DIR

IS_CI = True if os.getenv("CI") else False

SCALE_FACTOR = 0.2
NUM_PARTS = [1] if IS_CI else [1, 2]
SOURCE_TYPES = ["in-memory"] if IS_CI else ["parquet", "in-memory"]

memray_stats = defaultdict(dict)


def pytest_terminal_summary(terminalreporter):
    if memray_stats:
        for group, group_stats in sorted(memray_stats.items()):
            terminalreporter.write_sep("-", f"Memray Stats for Group: {group}")
            for nodeid, stats in group_stats.items():
                terminalreporter.write_line(
                    f"{nodeid} \t Peak Memory: {stats['peak_memory']} MB \t Total Allocations: {stats['total_allocations']}"
                )
                terminalreporter.ensure_newline()


@pytest.fixture
def benchmark_with_memray(request, benchmark):
    def track_mem(func, group):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_file_path = os.path.join(tmpdir, "memray_output.bin")
            with memray.Tracker(output_file_path):
                res = func()

            reader = memray.FileReader(output_file_path)
            stats = {
                "peak_memory": reader.metadata.peak_memory / 1024 / 1024,
                "total_allocations": reader.metadata.total_allocations,
            }
            memray_stats[group][request.node.nodeid] = stats

        return res

    def benchmark_wrapper(func, group):
        benchmark.group = group
        # If running in CI, just run the benchmark
        if os.getenv("CI"):
            return benchmark(func)
        else:
            benchmark(func)
            return track_mem(func, group)

    return benchmark_wrapper


@pytest.fixture(scope="session", params=NUM_PARTS)
def gen_tpch(request):
    # Parametrize the number of parts for each file so that we run tests on single-partition files and multi-partition files
    num_parts = request.param

    csv_files_location = data_generation.gen_csv_files(TPCH_DBGEN_DIR, num_parts, SCALE_FACTOR)
    parquet_files_location = data_generation.gen_parquet(csv_files_location)

    in_memory_tables = {}
    for tbl_name in data_generation.SCHEMA.keys():
        arrow_table = daft.read_parquet(f"{parquet_files_location}/{tbl_name}/*").to_arrow()
        in_memory_tables[tbl_name] = daft.from_arrow(arrow_table)

    sqlite_path = data_generation.gen_sqlite_db(
        csv_filepath=csv_files_location,
        num_parts=num_parts,
    )

    return (
        csv_files_location,
        parquet_files_location,
        in_memory_tables,
        num_parts,
    ), sqlite_path


@pytest.fixture(scope="module", params=SOURCE_TYPES)
def get_df(gen_tpch, request):
    (csv_files_location, parquet_files_location, in_memory_tables, num_parts), _ = gen_tpch
    source_type = request.param

    def _get_df(tbl_name: str):
        if source_type == "csv":
            local_fs = LocalFileSystem()
            nonchunked_filepath = f"{csv_files_location}/{tbl_name}.tbl"
            chunked_filepath = nonchunked_filepath + ".*"
            try:
                local_fs.expand_path(chunked_filepath)
                fp = chunked_filepath
            except FileNotFoundError:
                fp = nonchunked_filepath

            df = daft.read_csv(
                fp,
                has_headers=False,
                delimiter="|",
            )
            df = df.select(
                *[
                    daft.col(autoname).alias(colname)
                    for autoname, colname in zip(df.column_names, data_generation.SCHEMA[tbl_name])
                ]
            )
        elif source_type == "parquet":
            fp = f"{parquet_files_location}/{tbl_name}/*"
            df = daft.read_parquet(fp)
        elif source_type == "in-memory":
            df = in_memory_tables[tbl_name]

        return df

    return _get_df, num_parts



from __future__ import annotations

import dataclasses
import logging
import os
import pathlib
import sys
import urllib.parse
from datetime import datetime, timezone
from typing import Any

from daft.daft import FileFormat, FileInfos, IOConfig, io_glob
from daft.dependencies import fsspec, pafs
from daft.expressions.expressions import ExpressionsProjection, col
from daft.recordbatch import MicroPartition

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class PyArrowFSWithExpiry:
    fs: pafs.FileSystem
    expiry: datetime | None


_CACHED_FSES: dict[tuple[str, IOConfig | None], PyArrowFSWithExpiry] = {}


def _get_fs_from_cache(protocol: str, io_config: IOConfig | None) -> pafs.FileSystem | None:
    """Get an instantiated pyarrow filesystem from the cache based on the URI protocol.

    Returns None if no such cache entry exists.
    """
    global _CACHED_FSES

    if (protocol, io_config) in _CACHED_FSES:
        fs = _CACHED_FSES[(protocol, io_config)]

        if fs.expiry is None or fs.expiry > datetime.now(timezone.utc):
            return fs.fs

    return None


def _put_fs_in_cache(protocol: str, fs: pafs.FileSystem, io_config: IOConfig | None, expiry: datetime | None) -> None:
    """Put pyarrow filesystem in cache under provided protocol."""
    global _CACHED_FSES

    _CACHED_FSES[(protocol, io_config)] = PyArrowFSWithExpiry(fs, expiry)


def get_filesystem(protocol: str, **kwargs: Any) -> fsspec.AbstractFileSystem:
    if protocol == "s3" or protocol == "s3a":
        try:
            import botocore.session
        except ImportError:
            logger.error(
                "Error when importing botocore. install daft[aws] for the required 3rd party dependencies to interact with AWS S3 (https://docs.getdaft.io/en/latest/install)"
            )
            raise

        s3fs_kwargs = {}

        # If accessing S3 without credentials, use anonymous access: https://github.com/Eventual-Inc/Daft/issues/503
        credentials_available = botocore.session.get_session().get_credentials() is not None
        if not credentials_available:
            logger.warning(
                "AWS credentials not found - using anonymous access to S3 which will fail if the bucket you are accessing is not a public bucket. See boto3 documentation for more details on configuring your AWS credentials: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html"
            )
            s3fs_kwargs["anon"] = True

        kwargs = {**kwargs, **s3fs_kwargs}

    try:
        klass = fsspec.get_filesystem_class(protocol)
    except ImportError:
        logger.error(
            "Error when importing dependencies for accessing data with: %s. Please ensure that daft was installed with the appropriate extra dependencies (https://docs.getdaft.io/en/latest/install)",
            protocol,
        )
        raise

    fs = klass(**kwargs)
    return fs


def get_protocol_from_path(path: str) -> str:
    parsed_scheme = urllib.parse.urlparse(path, allow_fragments=False).scheme
    parsed_scheme = parsed_scheme.lower()
    if parsed_scheme == "" or parsed_scheme is None:
        return "file"
    if sys.platform == "win32" and len(parsed_scheme) == 1 and ("a" <= parsed_scheme) and (parsed_scheme <= "z"):
        return "file"
    return parsed_scheme


_CANONICAL_PROTOCOLS = {
    "gcs": "gs",
    "https": "http",
    "s3a": "s3",
    "ssh": "sftp",
    "arrow_hdfs": "hdfs",
    "az": "abfs",
    "blockcache": "cached",
    "jlab": "jupyter",
}


def canonicalize_protocol(protocol: str) -> str:
    """Return the canonical protocol from the provided protocol, such that there's a 1:1 mapping between protocols and pyarrow/fsspec filesystem implementations."""
    return _CANONICAL_PROTOCOLS.get(protocol, protocol)


def _resolve_paths_and_filesystem(
    paths: str | pathlib.Path | list[str],
    io_config: IOConfig | None = None,
) -> tuple[list[str], pafs.FileSystem]:
    """Resolves and normalizes the provided path and infers it's filesystem.

    Also ensures that the inferred filesystem is compatible with the passed filesystem, if provided.

    Args:
        paths: A single file/directory path or a list of file/directory paths.
            A list of paths can contain both files and directories.
        io_config: A Daft IOConfig that should be best-effort applied onto the returned
            FileSystem
    """
    if isinstance(paths, pathlib.Path):
        paths = str(paths)
    if isinstance(paths, str):
        paths = [paths]
    assert isinstance(paths, list), paths
    assert all(isinstance(p, str) for p in paths), paths
    assert len(paths) > 0, paths

    # Sanitize s3a/s3n protocols, which are produced by Hadoop-based systems as a way of denoting which s3
    # filesystem client to use. However this doesn't matter for Daft, and PyArrow cannot recognize these protocols.
    paths = [f"s3://{p[6:]}" if p.startswith("s3a://") or p.startswith("s3n://") else p for p in paths]

    # Ensure that protocols for all paths are consistent, i.e. that they would map to the
    # same filesystem.
    protocols = {get_protocol_from_path(path) for path in paths}
    canonicalized_protocols = {canonicalize_protocol(protocol) for protocol in protocols}
    if len(canonicalized_protocols) > 1:
        raise ValueError(
            "All paths must have the same canonical protocol to ensure that they are all "
            f"hitting the same storage backend, but got protocols {protocols} with canonical "
            f"protocols - {canonicalized_protocols} and full paths - {paths}"
        )

    # Canonical protocol shared by all paths.
    protocol = next(iter(canonicalized_protocols))

    # Try to get filesystem from protocol -> fs cache.
    resolved_filesystem = _get_fs_from_cache(protocol, io_config)
    if resolved_filesystem is None:
        # Resolve path and filesystem for the first path.
        # We use this first resolved filesystem for validation on all other paths.
        resolved_path, resolved_filesystem, expiry = _infer_filesystem(paths[0], io_config)

        # Put resolved filesystem in cache under these paths' canonical protocol.
        _put_fs_in_cache(protocol, resolved_filesystem, io_config, expiry)
    else:
        resolved_path = _validate_filesystem(paths[0], resolved_filesystem, io_config)

    # filesystem should be a non-None pyarrow FileSystem at this point, either
    # user-provided, taken from the cache, or inferred from the first path.
    assert resolved_filesystem is not None and isinstance(resolved_filesystem, pafs.FileSystem)

    # Resolve all other paths and validate with the user-provided/cached/inferred filesystem.
    resolved_paths = [resolved_path]
    for path in paths[1:]:
        resolved_path = _validate_filesystem(path, resolved_filesystem, io_config)
        resolved_paths.append(resolved_path)

    return resolved_paths, resolved_filesystem


def _validate_filesystem(path: str, fs: pafs.FileSystem, io_config: IOConfig | None) -> str:
    resolved_path, inferred_fs, _ = _infer_filesystem(path, io_config)
    if not isinstance(fs, type(inferred_fs)):
        raise RuntimeError(
            f"Cannot read multiple paths with different inferred PyArrow filesystems. Expected: {fs} but received: {inferred_fs}"
        )
    return resolved_path


def _infer_filesystem(
    path: str,
    io_config: IOConfig | None,
) -> tuple[str, pafs.FileSystem, datetime | None]:
    """Resolves and normalizes the provided path and infers its filesystem and expiry.

    Also ensures that the inferred filesystem is compatible with the passedfilesystem, if provided.

    Args:
        path: A single file/directory path.
        io_config: A Daft IOConfig that should be best-effort applied onto the returned
            FileSystem
    """
    protocol = get_protocol_from_path(path)
    translated_kwargs: dict[str, Any]

    def _set_if_not_none(kwargs: dict[str, Any], key: str, val: Any | None) -> None:
        """Helper method used when setting kwargs for pyarrow."""
        if val is not None:
            kwargs[key] = val

    ###
    # S3
    ###
    if protocol == "s3":
        translated_kwargs = {}
        if io_config is not None and io_config.s3 is not None:
            s3_config = io_config.s3
            _set_if_not_none(translated_kwargs, "endpoint_override", s3_config.endpoint_url)
            _set_if_not_none(translated_kwargs, "access_key", s3_config.key_id)
            _set_if_not_none(translated_kwargs, "secret_key", s3_config.access_key)
            _set_if_not_none(translated_kwargs, "session_token", s3_config.session_token)
            _set_if_not_none(translated_kwargs, "region", s3_config.region_name)
            _set_if_not_none(translated_kwargs, "anonymous", s3_config.anonymous)
            if s3_config.num_tries is not None:
                try:
                    from pyarrow.fs import AwsStandardS3RetryStrategy

                    translated_kwargs["retry_strategy"] = AwsStandardS3RetryStrategy(max_attempts=s3_config.num_tries)
                except ImportError:
                    pass  # Config does not exist in pyarrow 7.0.0

            expiry = None
            if (s3_creds := s3_config.provide_cached_credentials()) is not None:
                _set_if_not_none(translated_kwargs, "access_key", s3_creds.key_id)
                _set_if_not_none(translated_kwargs, "secret_key", s3_creds.access_key)
                _set_if_not_none(translated_kwargs, "session_token", s3_creds.session_token)

                expiry = s3_creds.expiry

        resolved_filesystem = pafs.S3FileSystem(**translated_kwargs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, expiry

    ###
    # Local
    ###
    elif protocol == "file":
        resolved_filesystem = pafs.LocalFileSystem()
        path = os.path.expanduser(path)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, None

    ###
    # GCS
    ###
    elif protocol in {"gs", "gcs"}:
        try:
            from pyarrow.fs import GcsFileSystem
        except ImportError:
            raise ImportError(
                "Unable to import GcsFileSystem from pyarrow - please ensure you have pyarrow >= 9.0 or consider "
                "using Daft's native GCS IO code instead"
            )

        translated_kwargs = {}
        if io_config is not None and io_config.gcs is not None:
            gcs_config = io_config.gcs
            _set_if_not_none(translated_kwargs, "anonymous", gcs_config.anonymous)
            _set_if_not_none(translated_kwargs, "project_id", gcs_config.project_id)

        resolved_filesystem = GcsFileSystem(**translated_kwargs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
        return resolved_path, resolved_filesystem, None

    ###
    # HTTP: Use FSSpec as a fallback
    ###
    elif protocol in {"http", "https"}:
        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)
        fsspec_fs = fsspec_fs_cls()
        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
        resolved_path = resolved_filesystem.normalize_path(resolved_path)
        return resolved_path, resolved_filesystem, None

    ###
    # Azure: Use FSSpec as a fallback
    ###
    elif protocol in {"az", "abfs", "abfss"}:
        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)

        if io_config is not None:
            # TODO: look into support for other AzureConfig parameters
            fsspec_fs = fsspec_fs_cls(
                account_name=io_config.azure.storage_account,
                account_key=io_config.azure.access_key,
                sas_token=io_config.azure.sas_token,
                tenant_id=io_config.azure.tenant_id,
                client_id=io_config.azure.client_id,
                client_secret=io_config.azure.client_secret,
                anon=io_config.azure.anonymous,
            )
        else:
            fsspec_fs = fsspec_fs_cls()
        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
        resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(resolved_path))
        return resolved_path, resolved_filesystem, None

    else:
        raise NotImplementedError(f"Cannot infer PyArrow filesystem for protocol {protocol}: please file an issue!")


def _unwrap_protocol(path: str) -> str:
    """Slice off any protocol prefixes on path."""
    parsed = urllib.parse.urlparse(path, allow_fragments=False)  # support '#' in path
    query = "?" + parsed.query if parsed.query else ""  # support '?' in path
    return parsed.netloc + parsed.path + query


###
# File globbing
###


def glob_path_with_stats(
    path: str,
    file_format: FileFormat | None,
    io_config: IOConfig | None,
) -> FileInfos:
    """Glob a path, returning a FileInfos."""
    files = io_glob(path, io_config=io_config)
    filepaths_to_infos = {f["path"]: {"size": f["size"], "type": f["type"]} for f in files}

    # Set number of rows if available.
    if file_format is not None and file_format == FileFormat.Parquet:
        parquet_statistics = MicroPartition.read_parquet_statistics(
            list(filepaths_to_infos.keys()),
            io_config=io_config,
        ).to_pydict()
        for path, num_rows in zip(parquet_statistics["uris"], parquet_statistics["row_count"]):
            filepaths_to_infos[path]["rows"] = num_rows

    file_paths = []
    file_sizes = []
    num_rows = []
    for path, infos in filepaths_to_infos.items():
        file_paths.append(path)
        file_sizes.append(infos.get("size"))
        num_rows.append(infos.get("rows"))

    return FileInfos.from_infos(file_paths=file_paths, file_sizes=file_sizes, num_rows=num_rows)


###
# Path joining
###


def join_path(fs: pafs.FileSystem, base_path: str, *sub_paths: str) -> str:
    """Join a base path with sub-paths using the appropriate path separator for the given filesystem."""
    if isinstance(fs, pafs.LocalFileSystem):
        return os.path.join(base_path, *sub_paths)
    else:
        return f"{base_path.rstrip('/')}/{'/'.join(sub_paths)}"


def overwrite_files(
    written_file_paths: list[str],
    root_dir: str | pathlib.Path,
    io_config: IOConfig | None,
    overwrite_partitions: bool,
) -> None:
    [resolved_path], fs = _resolve_paths_and_filesystem(root_dir, io_config=io_config)

    all_file_paths = []
    if overwrite_partitions:
        # Get all files in ONLY the directories that were written to.

        written_dirs = set(str(pathlib.Path(path).parent) for path in written_file_paths)
        for dir in written_dirs:
            file_selector = pafs.FileSelector(dir, recursive=True)
            try:
                all_file_paths.extend(
                    [info.path for info in fs.get_file_info(file_selector) if info.type == pafs.FileType.File]
                )
            except FileNotFoundError:
                continue
    else:
        # Get all files in the root directory.

        file_selector = pafs.FileSelector(resolved_path, recursive=True)
        try:
            all_file_paths.extend(
                [info.path for info in fs.get_file_info(file_selector) if info.type == pafs.FileType.File]
            )
        except FileNotFoundError:
            # The root directory does not exist, so there are no files to delete.
            return

    all_file_paths_df = MicroPartition.from_pydict({"path": all_file_paths})

    # Find the files that were not written to in this run and delete them.
    to_delete = all_file_paths_df.filter(ExpressionsProjection([~(col("path").is_in(written_file_paths))]))

    # TODO: Look into parallelizing this
    for entry in to_delete.get_column_by_name("path"):
        fs.delete_file(entry)



import subprocess
import time
import typing
from typing import Optional

import gha_run_cluster_job
from github import Auth, Github
from github.Workflow import Workflow
from github.WorkflowRun import WorkflowRun

RETRY_ATTEMPTS = 5

auth = Auth.Token(gha_run_cluster_job.get_oauth_token())
g = Github(auth=auth)
repo = g.get_repo("Eventual-Inc/Daft")


def dispatch(workflow: Workflow, branch_name: str, inputs: dict) -> WorkflowRun:
    pre_creation_latest_run = get_latest_run(workflow)

    print(f"Launching workflow '{workflow.name}' on the branch '{branch_name}'")
    created = workflow.create_dispatch(
        ref=branch_name,
        inputs=inputs,
    )
    if not created:
        raise RuntimeError("Could not create workflow, suggestion: run again with --verbose")

    post_creation_latest_run = None
    for _ in range(RETRY_ATTEMPTS):
        post_creation_latest_run = get_latest_run(workflow)
        if pre_creation_latest_run.run_number == post_creation_latest_run.run_number:
            sleep_and_then_retry()
        elif pre_creation_latest_run.run_number < post_creation_latest_run.run_number:
            break
        else:
            typing.assert_never(
                "Run numbers are always returned in sorted order and are always monotonically increasing"
            )
    if not post_creation_latest_run:
        raise RuntimeError(f"Unable to locate the new run request for the '{workflow.name}' workflow")

    print(f"Launched new '{workflow.name}' workflow with id: {post_creation_latest_run.id}")
    print(f"View the workflow run at: {post_creation_latest_run.html_url}")

    return post_creation_latest_run


def sleep_and_then_retry(sleep_amount_sec: int = 3):
    time.sleep(sleep_amount_sec)


def get_latest_run(workflow: Workflow) -> WorkflowRun:
    for _ in range(RETRY_ATTEMPTS):
        runs = workflow.get_runs()

        if runs.totalCount > 0:
            return runs[0]

        sleep_and_then_retry()

    raise RuntimeError("Unable to list all workflow invocations")


def get_name_and_commit_hash(local_branch_name: Optional[str]) -> tuple[str, str]:
    local_branch_name = local_branch_name or "HEAD"
    remote_branch_name = local_branch_name

    try:
        # Check if the branch has a remote tracking branch.
        local_branch_name = (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", f"{local_branch_name}@{{upstream}}"], stderr=subprocess.STDOUT
            )
            .strip()
            .decode("utf-8")
        )
        # Strip the upstream name from the branch to get the branch name on the remote repo.
        remote_branch_name = local_branch_name.split("/", 1)[1]
    except subprocess.CalledProcessError:
        local_branch_name = (
            subprocess.check_output(["git", "rev-parse", "--abbrev-ref", local_branch_name], stderr=subprocess.STDOUT)
            .strip()
            .decode("utf-8")
        )
        remote_branch_name = local_branch_name

    commit_hash = (
        subprocess.check_output(["git", "rev-parse", local_branch_name], stderr=subprocess.STDOUT)
        .strip()
        .decode("utf-8")
    )
    # Return the remote branch name for the github action.
    return remote_branch_name, commit_hash


def parse_questions(questions: Optional[str], total_number_of_questions: int) -> list[int]:
    if questions is None:
        return list(range(1, total_number_of_questions + 1))
    else:

        def to_int(q: str) -> int:
            question = int(q)
            if question > total_number_of_questions:
                raise ValueError(
                    f"Question number should be less than {total_number_of_questions}, instead got {question}"
                )
            return question

        return list(map(to_int, filter(lambda q: q, questions.split(","))))



"""Get a Daft DataFrame corresponding to a TPC-H question.

You may also run this file directly as such: `python answers_sql.py <path to TPC-H parquet data dir> <question number>`
"""

from __future__ import annotations

import os
import sys

import daft
from daft import col
from daft.io import IOConfig, S3Config
from daft.sql import SQLCatalog

TABLE_NAMES = [
    "part",
    "supplier",
    "partsupp",
    "customer",
    "orders",
    "lineitem",
    "nation",
    "region",
]


def lowercase_column_names(df):
    return df.select(*[col(name).alias(name.lower()) for name in df.column_names])


def q21(get_df) -> daft.DataFrame:
    supplier = get_df("supplier")
    nation = get_df("nation")
    lineitem = get_df("lineitem")
    orders = get_df("orders")

    res_1 = (
        lineitem.select("L_SUPPKEY", "L_ORDERKEY")
        # distinct not needed? either way we should stop using dataframe answers soon
        # .distinct()
        .groupby("L_ORDERKEY")
        .agg(col("L_SUPPKEY").count().alias("nunique_col"))
        .where(col("nunique_col") > 1)
        .join(lineitem.where(col("L_RECEIPTDATE") > col("L_COMMITDATE")), on="L_ORDERKEY")
    )

    daft_df = (
        res_1.select("L_SUPPKEY", "L_ORDERKEY")
        .groupby("L_ORDERKEY")
        .agg(col("L_SUPPKEY").count().alias("nunique_col"))
        .join(res_1, on="L_ORDERKEY")
        .join(supplier, left_on="L_SUPPKEY", right_on="S_SUPPKEY")
        .join(nation, left_on="S_NATIONKEY", right_on="N_NATIONKEY")
        .join(orders, left_on="L_ORDERKEY", right_on="O_ORDERKEY")
        .where((col("nunique_col") == 1) & (col("N_NAME") == "SAUDI ARABIA") & (col("O_ORDERSTATUS") == "F"))
        .groupby("S_NAME")
        .agg(col("O_ORDERKEY").count().alias("numwait"))
        .sort(["numwait", "S_NAME"], desc=[True, False])
        .limit(100)
    )

    return daft_df


def get_answer(q: int, get_df) -> daft.DataFrame:
    assert 1 <= q <= 22, f"TPC-H has 22 questions, received q={q}"

    if q == 21:
        # TODO: remove this once we support q21
        return q21(get_df)
    else:
        catalog = SQLCatalog({tbl: lowercase_column_names(get_df(tbl)) for tbl in TABLE_NAMES})

        module_dir = os.path.dirname(os.path.abspath(__file__))
        query_file_path = os.path.join(module_dir, f"queries/{q:02}.sql")

        with open(query_file_path) as query_file:
            query = query_file.read()
        return daft.sql(query, catalog=catalog)


def main(parquet_path, q):
    s3_config_from_env = S3Config.from_env()
    io_config = IOConfig(s3=s3_config_from_env)

    def get_df(name):
        return daft.read_parquet(f"{parquet_path}{name}/*", io_config=io_config)

    daft_df = get_answer(q, get_df)

    daft_df.collect()


if __name__ == "__main__":
    parquet_path = sys.argv[1]
    q = int(sys.argv[2])
    main(parquet_path, q)



from __future__ import annotations

import pytest

import daft


@pytest.mark.integration()
def test_url_download_gcs_public_special_characters(small_images_s3_paths):
    df = daft.from_glob_path("gs://daft-public-data-gs/test_naming/**")
    df = df.with_column("data", df["path"].url.download())

    assert df.to_pydict() == {
        "path": [
            "gs://daft-public-data-gs/test_naming/test. .txt",
            "gs://daft-public-data-gs/test_naming/test.%.txt",
            "gs://daft-public-data-gs/test_naming/test.-.txt",
            "gs://daft-public-data-gs/test_naming/test.=.txt",
            "gs://daft-public-data-gs/test_naming/test.?.txt",
        ],
        "size": [5, 5, 5, 5, 5],
        "num_rows": [None, None, None, None, None],
        "data": [b"test\n", b"test\n", b"test\n", b"test\n", b"test\n"],
    }



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from daft.daft import (
    DistributedPhysicalPlan,
    DistributedPhysicalPlanRunner,
    LocalPhysicalPlan,
    NativeExecutor,
    PyDaftExecutionConfig,
    RayPartitionRef,
    RaySwordfishTask,
    RaySwordfishWorker,
    RayTaskResult,
    set_compute_runtime_num_worker_threads,
)
from daft.recordbatch.micropartition import MicroPartition
from daft.runners.partitioning import (
    PartitionMetadata,
    PartitionSet,
)
from daft.runners.profiler import profile

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator, Iterator

    from daft.runners.ray_runner import RayMaterializedResult

try:
    import ray
except ImportError:
    raise

logger = logging.getLogger(__name__)


@ray.remote
class RaySwordfishActor:
    """RaySwordfishActor is a ray actor that runs local physical plans on swordfish.

    It is a stateless, async actor, and can run multiple plans concurrently and is able to retry itself and it's tasks.
    """

    def __init__(self, num_worker_threads: int) -> None:
        # Configure the number of worker threads for swordfish, according to the number of CPUs visible to ray.
        set_compute_runtime_num_worker_threads(num_worker_threads)
        self.native_executor = NativeExecutor()

    async def run_plan(
        self,
        plan: LocalPhysicalPlan,
        config: PyDaftExecutionConfig,
        psets: dict[str, list[ray.ObjectRef]],
        context: dict[str, str] | None,
    ) -> AsyncGenerator[MicroPartition | list[PartitionMetadata], None]:
        """Run a plan on swordfish and yield partitions."""
        with profile():
            psets = {k: await asyncio.gather(*v) for k, v in psets.items()}
            psets_mp = {k: [v._micropartition for v in v] for k, v in psets.items()}

            metas = []
            async for partition in self.native_executor.run_async(plan, psets_mp, config, None, context):
                if partition is None:
                    break
                mp = MicroPartition._from_pymicropartition(partition)
                metas.append(PartitionMetadata.from_table(mp))
                yield mp
            yield metas


@dataclass
class RaySwordfishTaskHandle:
    """RaySwordfishTaskHandle is a handle to a task that is running on a swordfish actor.

    It is used to asynchronously get the result of the task, cancel the task, and perform any post-task cleanup.
    """

    result_handle: ray.ObjectRef
    actor_handle: ray.actor.ActorHandle
    task: asyncio.Task[RayTaskResult] | None = None

    async def _get_result(self) -> RayTaskResult:
        try:
            await self.result_handle.completed()
            results = [result for result in self.result_handle]
            metadatas_ref = results.pop()

            metadatas = await metadatas_ref
            assert len(results) == len(metadatas)

            return RayTaskResult.success(
                [
                    RayPartitionRef(result, metadata.num_rows, metadata.size_bytes)
                    for result, metadata in zip(results, metadatas)
                ]
            )
        except (ray.exceptions.ActorDiedError, ray.exceptions.ActorUnschedulableError):
            return RayTaskResult.worker_died()
        except ray.exceptions.ActorUnavailableError:
            return RayTaskResult.worker_unavailable()
        except Exception as e:
            raise e

    async def get_result(self) -> RayTaskResult:
        self.task = asyncio.create_task(self._get_result())
        return await self.task

    def cancel(self) -> None:
        if self.task:
            self.task.cancel()
        ray.cancel(self.result_handle)


class RaySwordfishActorHandle:
    """RaySwordfishWorkerHandle is a wrapper around a ray swordfish actor.

    It is used to submit tasks to the worker and keep track of the worker's node id, handle, number of cpus, total and available memory.
    """

    def __init__(
        self,
        actor_handle: ray.actor.ActorHandle,
    ):
        self.actor_handle = actor_handle

    def submit_task(self, task: RaySwordfishTask) -> RaySwordfishTaskHandle:
        psets = {k: [v.object_ref for v in v] for k, v in task.psets().items()}
        result_handle = self.actor_handle.run_plan.options(name=task.name()).remote(
            task.plan(), task.config(), psets, task.context()
        )
        return RaySwordfishTaskHandle(
            result_handle,
            self.actor_handle,
        )

    def shutdown(self) -> None:
        ray.kill(self.actor_handle)


def start_ray_workers(existing_worker_ids: list[str]) -> list[RaySwordfishWorker]:
    handles = []
    for node in ray.nodes():
        if (
            "Resources" in node
            and "CPU" in node["Resources"]
            and "memory" in node["Resources"]
            and node["Resources"]["CPU"] > 0
            and node["Resources"]["memory"] > 0
            and node["NodeID"] not in existing_worker_ids
        ):
            actor = RaySwordfishActor.options(  # type: ignore
                scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
                    node_id=node["NodeID"],
                    soft=False,
                ),
            ).remote(num_worker_threads=int(node["Resources"]["CPU"]))
            actor_handle = RaySwordfishActorHandle(actor)
            handles.append(
                RaySwordfishWorker(
                    node["NodeID"],
                    actor_handle,
                    int(node["Resources"]["CPU"]),
                    int(node["Resources"].get("GPU", 0)),
                    int(node["Resources"]["memory"]),
                )
            )

    return handles


class FlotillaRunnerCore:
    """Core functionality for running distributed physical plans with Flotilla.

    This class contains the business logic for managing plans and their execution,
    separate from the Ray actor wrapper.
    """

    def __init__(self, on_actor: bool = False) -> None:
        self.curr_plans: dict[str, DistributedPhysicalPlan] = {}
        self.curr_result_gens: dict[str, AsyncIterator[RayPartitionRef]] = {}
        self.plan_runner = DistributedPhysicalPlanRunner(on_actor)

    def run_plan(
        self,
        plan: DistributedPhysicalPlan,
        partition_sets: dict[str, PartitionSet[ray.ObjectRef]],
    ) -> None:
        psets = {
            k: [RayPartitionRef(v.partition(), v.metadata().num_rows, v.metadata().size_bytes or 0) for v in v.values()]
            for k, v in partition_sets.items()
        }
        self.curr_plans[plan.id()] = plan
        self.curr_result_gens[plan.id()] = self.plan_runner.run_plan(plan, psets)

    async def get_next_partition(self, plan_id: str) -> RayMaterializedResult | None:
        from daft.runners.ray_runner import (
            PartitionMetadataAccessor,
            RayMaterializedResult,
        )

        if plan_id not in self.curr_result_gens:
            raise ValueError(f"Plan {plan_id} not found in FlotillaPlanRunner")

        next_partition_ref = await self.curr_result_gens[plan_id].__anext__()
        if next_partition_ref is None:
            self.curr_plans.pop(plan_id)
            self.curr_result_gens.pop(plan_id)
            return None

        metadata_accessor = PartitionMetadataAccessor.from_metadata_list(
            [PartitionMetadata(next_partition_ref.num_rows, next_partition_ref.size_bytes)]
        )
        materialized_result = RayMaterializedResult(
            partition=next_partition_ref.object_ref,
            metadatas=metadata_accessor,
            metadata_idx=0,
        )
        return materialized_result


class LocalFlotillaRunner:
    """Local wrapper around FlotillaPlanRunnerCore.

    This wrapper provides the same interface as FlotillaPlanRunner but without
    Ray actor overhead, useful for local testing or when distributed execution
    is not needed.
    """

    def __init__(self) -> None:
        self.loop = asyncio.new_event_loop()
        self.core = self.loop.run_until_complete(self._make_runner())

    async def _make_runner(self) -> FlotillaRunnerCore:
        return FlotillaRunnerCore(on_actor=False)

    def run_plan(
        self,
        plan: DistributedPhysicalPlan,
        partition_sets: dict[str, PartitionSet[ray.ObjectRef]],
    ) -> None:
        self.core.run_plan(plan, partition_sets)

    def get_next_partition(self, plan_id: str) -> RayMaterializedResult | None:
        """Synchronous version of get_next_partition that internally uses asyncio."""
        return self.loop.run_until_complete(self.core.get_next_partition(plan_id))


@ray.remote(
    num_cpus=0,
)
class RemoteFlotillaRunner:
    """Ray actor wrapper around FlotillaPlanRunnerCore.

    This actor provides the distributed interface for running plans,
    while delegating the actual work to the core class.
    """

    def __init__(self) -> None:
        self.core = FlotillaRunnerCore(on_actor=True)

    def run_plan(
        self,
        plan: DistributedPhysicalPlan,
        partition_sets: dict[str, PartitionSet[ray.ObjectRef]],
    ) -> None:
        self.core.run_plan(plan, partition_sets)

    async def get_next_partition(self, plan_id: str) -> RayMaterializedResult | None:
        return await self.core.get_next_partition(plan_id)


FLOTILLA_RUNER_NAMESPACE = "daft"
FLOTILLA_RUNNER_NAME = "flotilla-plan-runner"


def get_head_node_id() -> str | None:
    for node in ray.nodes():
        if (
            "Resources" in node
            and "node:__internal_head__" in node["Resources"]
            and node["Resources"]["node:__internal_head__"] == 1
        ):
            return node["NodeID"]
    return None


class FlotillaRunner:
    """FlotillaRunner is a wrapper around FlotillaRunnerCore that provides a Ray actor interface."""

    def __init__(self, use_actor: bool = False) -> None:
        self.use_actor = use_actor
        if self.use_actor:
            head_node_id = get_head_node_id()
            self.runner = RemoteFlotillaRunner.options(  # type: ignore
                name=FLOTILLA_RUNNER_NAME,
                namespace=FLOTILLA_RUNER_NAMESPACE,
                get_if_exists=True,
                scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
                    node_id=head_node_id,
                    soft=False,
                )
                if head_node_id is not None
                else "DEFAULT",
            ).remote()
        else:
            self.runner = LocalFlotillaRunner()

    def stream_plan(
        self, plan: DistributedPhysicalPlan, partition_sets: dict[str, PartitionSet[ray.ObjectRef]]
    ) -> Iterator[RayMaterializedResult]:
        plan_id = plan.id()
        if self.use_actor:
            ray.get(self.runner.run_plan.remote(plan, partition_sets))
            while True:
                materialized_result = ray.get(self.runner.get_next_partition.remote(plan_id))
                if materialized_result is None:
                    break
                yield materialized_result
        else:
            self.runner.run_plan(plan, partition_sets)
            while True:
                materialized_result = self.runner.get_next_partition(plan_id)
                if materialized_result is None:
                    break
                yield materialized_result



from __future__ import annotations

import datetime
from typing import Callable

from daft import DataFrame, col, lit

GetDFFunc = Callable[[str], DataFrame]


def q1(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")

    discounted_price = col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT"))
    taxed_discounted_price = discounted_price * (1 + col("L_TAX"))
    daft_df = (
        lineitem.where(col("L_SHIPDATE") <= datetime.date(1998, 9, 2))
        .groupby(col("L_RETURNFLAG"), col("L_LINESTATUS"))
        .agg(
            col("L_QUANTITY").sum().alias("sum_qty"),
            col("L_EXTENDEDPRICE").sum().alias("sum_base_price"),
            discounted_price.sum().alias("sum_disc_price"),
            taxed_discounted_price.sum().alias("sum_charge"),
            col("L_QUANTITY").mean().alias("avg_qty"),
            col("L_EXTENDEDPRICE").mean().alias("avg_price"),
            col("L_DISCOUNT").mean().alias("avg_disc"),
            col("L_QUANTITY").count().alias("count_order"),
        )
        .sort(["L_RETURNFLAG", "L_LINESTATUS"])
    )
    return daft_df


def q2(get_df: GetDFFunc) -> DataFrame:
    region = get_df("region")
    nation = get_df("nation")
    supplier = get_df("supplier")
    partsupp = get_df("partsupp")
    part = get_df("part")

    europe = (
        region.where(col("R_NAME") == "EUROPE")
        .join(nation, left_on=col("R_REGIONKEY"), right_on=col("N_REGIONKEY"))
        .join(supplier, left_on=col("N_NATIONKEY"), right_on=col("S_NATIONKEY"))
        .join(partsupp, left_on=col("S_SUPPKEY"), right_on=col("PS_SUPPKEY"))
    )

    brass = part.where((col("P_SIZE") == 15) & col("P_TYPE").str.endswith("BRASS")).join(
        europe,
        left_on=col("P_PARTKEY"),
        right_on=col("PS_PARTKEY"),
    )
    min_cost = brass.groupby(col("P_PARTKEY")).agg(col("PS_SUPPLYCOST").min().alias("min"))

    daft_df = (
        brass.join(min_cost, on=col("P_PARTKEY"))
        .where(col("PS_SUPPLYCOST") == col("min"))
        .select(
            col("S_ACCTBAL"),
            col("S_NAME"),
            col("N_NAME"),
            col("P_PARTKEY"),
            col("P_MFGR"),
            col("S_ADDRESS"),
            col("S_PHONE"),
            col("S_COMMENT"),
        )
        .sort(by=["S_ACCTBAL", "N_NAME", "S_NAME", "P_PARTKEY"], desc=[True, False, False, False])
        .limit(100)
    )
    return daft_df


def q3(get_df: GetDFFunc) -> DataFrame:
    def decrease(x, y):
        return x * (1 - y)

    customer = get_df("customer").where(col("C_MKTSEGMENT") == "BUILDING")
    orders = get_df("orders").where(col("O_ORDERDATE") < datetime.date(1995, 3, 15))
    lineitem = get_df("lineitem").where(col("L_SHIPDATE") > datetime.date(1995, 3, 15))

    daft_df = (
        customer.join(orders, left_on=col("C_CUSTKEY"), right_on=col("O_CUSTKEY"))
        .select(col("O_ORDERKEY"), col("O_ORDERDATE"), col("O_SHIPPRIORITY"))
        .join(lineitem, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .select(
            col("O_ORDERKEY"),
            decrease(col("L_EXTENDEDPRICE"), col("L_DISCOUNT")).alias("volume"),
            col("O_ORDERDATE"),
            col("O_SHIPPRIORITY"),
        )
        .groupby(col("O_ORDERKEY"), col("O_ORDERDATE"), col("O_SHIPPRIORITY"))
        .agg(col("volume").sum().alias("revenue"))
        .sort(by=["revenue", "O_ORDERDATE"], desc=[True, False])
        .limit(10)
        .select("O_ORDERKEY", "revenue", "O_ORDERDATE", "O_SHIPPRIORITY")
    )
    return daft_df


def q4(get_df: GetDFFunc) -> DataFrame:
    orders = get_df("orders")
    lineitems = get_df("lineitem")

    orders = orders.where(
        (col("O_ORDERDATE") >= datetime.date(1993, 7, 1)) & (col("O_ORDERDATE") < datetime.date(1993, 10, 1))
    )

    lineitems = lineitems.where(col("L_COMMITDATE") < col("L_RECEIPTDATE"))

    daft_df = (
        orders.join(lineitems, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"), how="semi")
        .groupby(col("O_ORDERPRIORITY"))
        .agg(col("O_ORDERKEY").count().alias("order_count"))
        .sort(col("O_ORDERPRIORITY"))
    )
    return daft_df


def q5(get_df: GetDFFunc) -> DataFrame:
    orders = get_df("orders").where(
        (col("O_ORDERDATE") >= datetime.date(1994, 1, 1)) & (col("O_ORDERDATE") < datetime.date(1995, 1, 1))
    )
    region = get_df("region").where(col("R_NAME") == "ASIA")
    nation = get_df("nation")
    supplier = get_df("supplier")
    lineitem = get_df("lineitem")
    customer = get_df("customer")

    daft_df = (
        region.join(nation, left_on=col("R_REGIONKEY"), right_on=col("N_REGIONKEY"))
        .join(supplier, left_on=col("N_NATIONKEY"), right_on=col("S_NATIONKEY"))
        .join(lineitem, left_on=col("S_SUPPKEY"), right_on=col("L_SUPPKEY"))
        .select(col("N_NAME"), col("L_EXTENDEDPRICE"), col("L_DISCOUNT"), col("L_ORDERKEY"), col("N_NATIONKEY"))
        .join(orders, left_on=col("L_ORDERKEY"), right_on=col("O_ORDERKEY"))
        .join(customer, left_on=[col("O_CUSTKEY"), col("N_NATIONKEY")], right_on=[col("C_CUSTKEY"), col("C_NATIONKEY")])
        .select(col("N_NAME"), (col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT"))).alias("value"))
        .groupby(col("N_NAME"))
        .agg(col("value").sum().alias("revenue"))
        .sort(col("revenue"), desc=True)
    )
    return daft_df


def q6(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")
    daft_df = lineitem.where(
        (col("L_SHIPDATE") >= datetime.date(1994, 1, 1))
        & (col("L_SHIPDATE") < datetime.date(1995, 1, 1))
        & (col("L_DISCOUNT") >= 0.05)
        & (col("L_DISCOUNT") <= 0.07)
        & (col("L_QUANTITY") < 24)
    ).sum(col("L_EXTENDEDPRICE") * col("L_DISCOUNT"))
    return daft_df


def q7(get_df: GetDFFunc) -> DataFrame:
    def decrease(x, y):
        return x * (1 - y)

    lineitem = get_df("lineitem").where(
        (col("L_SHIPDATE") >= datetime.date(1995, 1, 1)) & (col("L_SHIPDATE") <= datetime.date(1996, 12, 31))
    )
    nation = get_df("nation").where((col("N_NAME") == "FRANCE") | (col("N_NAME") == "GERMANY"))
    supplier = get_df("supplier")
    customer = get_df("customer")
    orders = get_df("orders")

    supNation = (
        nation.join(supplier, left_on=col("N_NATIONKEY"), right_on=col("S_NATIONKEY"))
        .join(lineitem, left_on=col("S_SUPPKEY"), right_on=col("L_SUPPKEY"))
        .select(
            col("N_NAME").alias("supp_nation"),
            col("L_ORDERKEY"),
            col("L_EXTENDEDPRICE"),
            col("L_DISCOUNT"),
            col("L_SHIPDATE"),
        )
    )

    daft_df = (
        nation.join(customer, left_on=col("N_NATIONKEY"), right_on=col("C_NATIONKEY"))
        .join(orders, left_on=col("C_CUSTKEY"), right_on=col("O_CUSTKEY"))
        .select(col("N_NAME").alias("cust_nation"), col("O_ORDERKEY"))
        .join(supNation, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .where(
            ((col("supp_nation") == "FRANCE") & (col("cust_nation") == "GERMANY"))
            | ((col("supp_nation") == "GERMANY") & (col("cust_nation") == "FRANCE"))
        )
        .select(
            col("supp_nation"),
            col("cust_nation"),
            col("L_SHIPDATE").dt.year().alias("l_year"),
            decrease(col("L_EXTENDEDPRICE"), col("L_DISCOUNT")).alias("volume"),
        )
        .groupby(col("supp_nation"), col("cust_nation"), col("l_year"))
        .agg(col("volume").sum().alias("revenue"))
        .sort(by=["supp_nation", "cust_nation", "l_year"])
    )
    return daft_df


def q8(get_df: GetDFFunc) -> DataFrame:
    def decrease(x, y):
        return x * (1 - y)

    region = get_df("region").where(col("R_NAME") == "AMERICA")
    orders = get_df("orders").where(
        (col("O_ORDERDATE") <= datetime.date(1996, 12, 31)) & (col("O_ORDERDATE") >= datetime.date(1995, 1, 1))
    )
    part = get_df("part").where(col("P_TYPE") == "ECONOMY ANODIZED STEEL")
    nation = get_df("nation")
    supplier = get_df("supplier")
    lineitem = get_df("lineitem")
    customer = get_df("customer")

    nat = nation.join(supplier, left_on=col("N_NATIONKEY"), right_on=col("S_NATIONKEY"))

    line = (
        lineitem.select(
            col("L_PARTKEY"),
            col("L_SUPPKEY"),
            col("L_ORDERKEY"),
            decrease(col("L_EXTENDEDPRICE"), col("L_DISCOUNT")).alias("volume"),
        )
        .join(part, left_on=col("L_PARTKEY"), right_on=col("P_PARTKEY"))
        .join(nat, left_on=col("L_SUPPKEY"), right_on=col("S_SUPPKEY"))
    )

    daft_df = (
        nation.join(region, left_on=col("N_REGIONKEY"), right_on=col("R_REGIONKEY"))
        .select(col("N_NATIONKEY"))
        .join(customer, left_on=col("N_NATIONKEY"), right_on=col("C_NATIONKEY"))
        .select(col("C_CUSTKEY"))
        .join(orders, left_on=col("C_CUSTKEY"), right_on=col("O_CUSTKEY"))
        .select(col("O_ORDERKEY"), col("O_ORDERDATE"))
        .join(line, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .select(
            col("O_ORDERDATE").dt.year().alias("o_year"),
            col("volume"),
            (col("N_NAME") == "BRAZIL").if_else(col("volume"), 0.0).alias("case_volume"),
        )
        .groupby(col("o_year"))
        .agg(col("case_volume").sum().alias("case_volume_sum"), col("volume").sum().alias("volume_sum"))
        .select(col("o_year"), col("case_volume_sum") / col("volume_sum"))
        .sort(col("o_year"))
    )

    return daft_df


def q9(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")
    part = get_df("part")
    nation = get_df("nation")
    supplier = get_df("supplier")
    partsupp = get_df("partsupp")
    orders = get_df("orders")

    def expr(x, y, v, w):
        return x * (1 - y) - (v * w)

    linepart = part.where(col("P_NAME").str.contains("green")).join(
        lineitem, left_on=col("P_PARTKEY"), right_on=col("L_PARTKEY")
    )
    natsup = nation.join(supplier, left_on=col("N_NATIONKEY"), right_on=col("S_NATIONKEY"))

    daft_df = (
        linepart.join(natsup, left_on=col("L_SUPPKEY"), right_on=col("S_SUPPKEY"))
        .join(partsupp, left_on=[col("L_SUPPKEY"), col("P_PARTKEY")], right_on=[col("PS_SUPPKEY"), col("PS_PARTKEY")])
        .join(orders, left_on=col("L_ORDERKEY"), right_on=col("O_ORDERKEY"))
        .select(
            col("N_NAME"),
            col("O_ORDERDATE").dt.year().alias("o_year"),
            expr(col("L_EXTENDEDPRICE"), col("L_DISCOUNT"), col("PS_SUPPLYCOST"), col("L_QUANTITY")).alias("amount"),
        )
        .groupby(col("N_NAME"), col("o_year"))
        .agg(col("amount").sum())
        .sort(by=["N_NAME", "o_year"], desc=[False, True])
    )

    return daft_df


def q10(get_df: GetDFFunc) -> DataFrame:
    def decrease(x, y):
        return x * (1 - y)

    lineitem = get_df("lineitem").where(col("L_RETURNFLAG") == "R")
    orders = get_df("orders")
    nation = get_df("nation")
    customer = get_df("customer")

    daft_df = (
        orders.where(
            (col("O_ORDERDATE") < datetime.date(1994, 1, 1)) & (col("O_ORDERDATE") >= datetime.date(1993, 10, 1))
        )
        .join(customer, left_on=col("O_CUSTKEY"), right_on=col("C_CUSTKEY"))
        .join(nation, left_on=col("C_NATIONKEY"), right_on=col("N_NATIONKEY"))
        .join(lineitem, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .select(
            col("O_CUSTKEY"),
            col("C_NAME"),
            decrease(col("L_EXTENDEDPRICE"), col("L_DISCOUNT")).alias("volume"),
            col("C_ACCTBAL"),
            col("N_NAME"),
            col("C_ADDRESS"),
            col("C_PHONE"),
            col("C_COMMENT"),
        )
        .groupby(
            col("O_CUSTKEY"),
            col("C_NAME"),
            col("C_ACCTBAL"),
            col("C_PHONE"),
            col("N_NAME"),
            col("C_ADDRESS"),
            col("C_COMMENT"),
        )
        .agg(col("volume").sum().alias("revenue"))
        .sort(col("revenue"), desc=True)
        .select(
            col("O_CUSTKEY"),
            col("C_NAME"),
            col("revenue"),
            col("C_ACCTBAL"),
            col("N_NAME"),
            col("C_ADDRESS"),
            col("C_PHONE"),
            col("C_COMMENT"),
        )
        .limit(20)
    )

    return daft_df


def q11(get_df: GetDFFunc) -> DataFrame:
    partsupp = get_df("partsupp")
    supplier = get_df("supplier")
    nation = get_df("nation")

    var_1 = "GERMANY"
    var_2 = 0.0001 / 1

    res_1 = (
        partsupp.join(supplier, left_on=col("PS_SUPPKEY"), right_on=col("S_SUPPKEY"))
        .join(nation, left_on=col("S_NATIONKEY"), right_on=col("N_NATIONKEY"))
        .where(col("N_NAME") == var_1)
    )

    res_2 = res_1.agg((col("PS_SUPPLYCOST") * col("PS_AVAILQTY")).sum().alias("tmp")).select(
        col("tmp") * var_2, lit(1).alias("lit")
    )

    daft_df = (
        res_1.groupby("PS_PARTKEY")
        .agg(
            (col("PS_SUPPLYCOST") * col("PS_AVAILQTY")).sum().alias("value"),
        )
        .with_column("lit", lit(1))
        .join(res_2, on="lit")
        .where(col("value") > col("tmp"))
        .select(col("PS_PARTKEY"), col("value").round(2))
        .sort(col("value"), desc=True)
    )

    return daft_df


def q12(get_df: GetDFFunc) -> DataFrame:
    orders = get_df("orders")
    lineitem = get_df("lineitem")

    daft_df = (
        orders.join(lineitem, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .where(
            col("L_SHIPMODE").is_in(["MAIL", "SHIP"])
            & (col("L_COMMITDATE") < col("L_RECEIPTDATE"))
            & (col("L_SHIPDATE") < col("L_COMMITDATE"))
            & (col("L_RECEIPTDATE") >= datetime.date(1994, 1, 1))
            & (col("L_RECEIPTDATE") < datetime.date(1995, 1, 1))
        )
        .groupby(col("L_SHIPMODE"))
        .agg(
            col("O_ORDERPRIORITY").is_in(["1-URGENT", "2-HIGH"]).if_else(1, 0).sum().alias("high_line_count"),
            (~col("O_ORDERPRIORITY").is_in(["1-URGENT", "2-HIGH"])).if_else(1, 0).sum().alias("low_line_count"),
        )
        .sort(col("L_SHIPMODE"))
    )

    return daft_df


def q13(get_df: GetDFFunc) -> DataFrame:
    customers = get_df("customer")
    orders = get_df("orders")

    daft_df = (
        customers.join(
            orders.where(~col("O_COMMENT").str.match(".*special.*requests.*")),
            left_on="C_CUSTKEY",
            right_on="O_CUSTKEY",
            how="left",
        )
        .groupby(col("C_CUSTKEY"))
        .agg(col("O_ORDERKEY").count().alias("c_count"))
        .sort("C_CUSTKEY")
        .groupby("c_count")
        .agg(col("c_count").count().alias("custdist"))
        .sort(["custdist", "c_count"], desc=[True, True])
    )

    return daft_df


def q14(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")
    part = get_df("part")

    daft_df = (
        lineitem.join(part, left_on=col("L_PARTKEY"), right_on=col("P_PARTKEY"))
        .where((col("L_SHIPDATE") >= datetime.date(1995, 9, 1)) & (col("L_SHIPDATE") < datetime.date(1995, 10, 1)))
        .agg(
            col("P_TYPE")
            .str.startswith("PROMO")
            .if_else(col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT")), 0)
            .sum()
            .alias("tmp_1"),
            (col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT"))).sum().alias("tmp_2"),
        )
        .select(100.00 * (col("tmp_1") / col("tmp_2")).alias("promo_revenue"))
    )

    return daft_df


def q15(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")

    revenue = (
        lineitem.where(
            (col("L_SHIPDATE") >= datetime.date(1996, 1, 1)) & (col("L_SHIPDATE") < datetime.date(1996, 4, 1))
        )
        .groupby(col("L_SUPPKEY"))
        .agg((col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT"))).sum().alias("total_revenue"))
        .select(col("L_SUPPKEY").alias("supplier_no"), "total_revenue")
    )

    revenue = revenue.join(revenue.max("total_revenue"), on="total_revenue")

    supplier = get_df("supplier")

    daft_df = (
        supplier.join(revenue, left_on=col("S_SUPPKEY"), right_on=col("supplier_no"))
        .select("S_SUPPKEY", "S_NAME", "S_ADDRESS", "S_PHONE", "total_revenue")
        .sort("S_SUPPKEY")
    )

    return daft_df


def q16(get_df: GetDFFunc) -> DataFrame:
    part = get_df("part")
    partsupp = get_df("partsupp")

    supplier = get_df("supplier")

    suppkeys = supplier.where(col("S_COMMENT").str.match(".*Customer.*Complaints.*")).select(
        col("S_SUPPKEY"), col("S_SUPPKEY").alias("PS_SUPPKEY_RIGHT")
    )

    daft_df = (
        part.join(partsupp, left_on=col("P_PARTKEY"), right_on=col("PS_PARTKEY"))
        .where(
            (col("P_BRAND") != "Brand#45")
            & ~col("P_TYPE").str.startswith("MEDIUM POLISHED")
            & (col("P_SIZE").is_in([49, 14, 23, 45, 19, 3, 36, 9]))
        )
        .join(suppkeys, left_on="PS_SUPPKEY", right_on="S_SUPPKEY", how="left")
        .where(col("PS_SUPPKEY_RIGHT").is_null())
        .select("P_BRAND", "P_TYPE", "P_SIZE", "PS_SUPPKEY")
        .distinct()
        .groupby("P_BRAND", "P_TYPE", "P_SIZE")
        .agg(col("PS_SUPPKEY").count().alias("supplier_cnt"))
        .sort(["supplier_cnt", "P_BRAND", "P_TYPE", "P_SIZE"], desc=[True, False, False, False])
    )

    return daft_df


def q17(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")
    part = get_df("part")

    res_1 = part.where((col("P_BRAND") == "Brand#23") & (col("P_CONTAINER") == "MED BOX")).join(
        lineitem, left_on="P_PARTKEY", right_on="L_PARTKEY", how="left"
    )

    daft_df = (
        res_1.groupby("P_PARTKEY")
        .agg((0.2 * col("L_QUANTITY")).mean().alias("avg_quantity"))
        .select(col("P_PARTKEY").alias("key"), col("avg_quantity"))
        .join(res_1, left_on="key", right_on="P_PARTKEY")
        .where(col("L_QUANTITY") < col("avg_quantity"))
        .agg((col("L_EXTENDEDPRICE") / 7.0).sum().alias("avg_yearly"))
    )

    return daft_df


def q18(get_df: GetDFFunc) -> DataFrame:
    customer = get_df("customer")
    orders = get_df("orders")
    lineitem = get_df("lineitem")

    res_1 = lineitem.groupby("L_ORDERKEY").agg(col("L_QUANTITY").sum().alias("sum_qty")).where(col("sum_qty") > 300)

    daft_df = (
        orders.join(res_1, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .join(customer, left_on=col("O_CUSTKEY"), right_on=col("C_CUSTKEY"))
        .join(lineitem, left_on=col("O_ORDERKEY"), right_on=col("L_ORDERKEY"))
        .groupby("C_NAME", "C_CUSTKEY", "O_ORDERKEY", "O_ORDERDATE", "O_TOTALPRICE")
        .agg(col("L_QUANTITY").sum().alias("sum"))
        .select("C_NAME", "C_CUSTKEY", "O_ORDERKEY", col("O_ORDERDATE").alias("O_ORDERDAT"), "O_TOTALPRICE", "sum")
        .sort(["O_TOTALPRICE", "O_ORDERDAT"], desc=[True, False])
        .limit(100)
    )

    return daft_df


def q19(get_df: GetDFFunc) -> DataFrame:
    lineitem = get_df("lineitem")
    part = get_df("part")

    daft_df = (
        lineitem.join(part, left_on=col("L_PARTKEY"), right_on=col("P_PARTKEY"))
        .where(
            (
                (col("P_BRAND") == "Brand#12")
                & col("P_CONTAINER").is_in(["SM CASE", "SM BOX", "SM PACK", "SM PKG"])
                & (col("L_QUANTITY") >= 1)
                & (col("L_QUANTITY") <= 11)
                & (col("P_SIZE") >= 1)
                & (col("P_SIZE") <= 5)
                & col("L_SHIPMODE").is_in(["AIR", "AIR REG"])
                & (col("L_SHIPINSTRUCT") == "DELIVER IN PERSON")
            )
            | (
                (col("P_BRAND") == "Brand#23")
                & col("P_CONTAINER").is_in(["MED BAG", "MED BOX", "MED PKG", "MED PACK"])
                & (col("L_QUANTITY") >= 10)
                & (col("L_QUANTITY") <= 20)
                & (col("P_SIZE") >= 1)
                & (col("P_SIZE") <= 10)
                & col("L_SHIPMODE").is_in(["AIR", "AIR REG"])
                & (col("L_SHIPINSTRUCT") == "DELIVER IN PERSON")
            )
            | (
                (col("P_BRAND") == "Brand#34")
                & col("P_CONTAINER").is_in(["LG CASE", "LG BOX", "LG PACK", "LG PKG"])
                & (col("L_QUANTITY") >= 20)
                & (col("L_QUANTITY") <= 30)
                & (col("P_SIZE") >= 1)
                & (col("P_SIZE") <= 15)
                & col("L_SHIPMODE").is_in(["AIR", "AIR REG"])
                & (col("L_SHIPINSTRUCT") == "DELIVER IN PERSON")
            )
        )
        .agg((col("L_EXTENDEDPRICE") * (1 - col("L_DISCOUNT"))).sum().alias("revenue"))
    )

    return daft_df


def q20(get_df: GetDFFunc) -> DataFrame:
    supplier = get_df("supplier")
    nation = get_df("nation")
    part = get_df("part")
    partsupp = get_df("partsupp")
    lineitem = get_df("lineitem")

    res_1 = (
        lineitem.where(
            (col("L_SHIPDATE") >= datetime.date(1994, 1, 1)) & (col("L_SHIPDATE") < datetime.date(1995, 1, 1))
        )
        .groupby("L_PARTKEY", "L_SUPPKEY")
        .agg(((col("L_QUANTITY") * 0.5).sum()).alias("sum_quantity"))
    )

    res_2 = nation.where(col("N_NAME") == "CANADA")
    res_3 = supplier.join(res_2, left_on="S_NATIONKEY", right_on="N_NATIONKEY")

    daft_df = (
        part.where(col("P_NAME").str.startswith("forest"))
        .select("P_PARTKEY")
        .distinct()
        .join(partsupp, left_on="P_PARTKEY", right_on="PS_PARTKEY")
        .join(
            res_1,
            left_on=["PS_SUPPKEY", "P_PARTKEY"],
            right_on=["L_SUPPKEY", "L_PARTKEY"],
        )
        .where(col("PS_AVAILQTY") > col("sum_quantity"))
        .select("PS_SUPPKEY")
        .distinct()
        .join(res_3, left_on="PS_SUPPKEY", right_on="S_SUPPKEY")
        .select("S_NAME", "S_ADDRESS")
        .sort("S_NAME")
    )

    return daft_df


def q21(get_df: GetDFFunc) -> DataFrame:
    supplier = get_df("supplier")
    nation = get_df("nation")
    lineitem = get_df("lineitem")
    orders = get_df("orders")

    res_1 = (
        lineitem.select("L_SUPPKEY", "L_ORDERKEY")
        # distinct not needed? either way we should stop using dataframe answers soon
        # .distinct()
        .groupby("L_ORDERKEY")
        .agg(col("L_SUPPKEY").count().alias("nunique_col"))
        .where(col("nunique_col") > 1)
        .join(lineitem.where(col("L_RECEIPTDATE") > col("L_COMMITDATE")), on="L_ORDERKEY")
    )

    daft_df = (
        res_1.select("L_SUPPKEY", "L_ORDERKEY")
        .groupby("L_ORDERKEY")
        .agg(col("L_SUPPKEY").count().alias("nunique_col"))
        .join(res_1, on="L_ORDERKEY")
        .join(supplier, left_on="L_SUPPKEY", right_on="S_SUPPKEY")
        .join(nation, left_on="S_NATIONKEY", right_on="N_NATIONKEY")
        .join(orders, left_on="L_ORDERKEY", right_on="O_ORDERKEY")
        .where((col("nunique_col") == 1) & (col("N_NAME") == "SAUDI ARABIA") & (col("O_ORDERSTATUS") == "F"))
        .groupby("S_NAME")
        .agg(col("O_ORDERKEY").count().alias("numwait"))
        .sort(["numwait", "S_NAME"], desc=[True, False])
        .limit(100)
    )

    return daft_df


def q22(get_df: GetDFFunc) -> DataFrame:
    orders = get_df("orders")
    customer = get_df("customer")

    res_1 = (
        customer.with_column("cntrycode", col("C_PHONE").str.left(2))
        .where(col("cntrycode").is_in(["13", "31", "23", "29", "30", "18", "17"]))
        .select("C_ACCTBAL", "C_CUSTKEY", "cntrycode")
    )

    res_2 = (
        res_1.where(col("C_ACCTBAL") > 0).agg(col("C_ACCTBAL").mean().alias("avg_acctbal")).with_column("lit", lit(1))
    )

    daft_df = (
        res_1.join(orders, left_on="C_CUSTKEY", right_on="O_CUSTKEY", how="anti")
        .with_column("lit", lit(1))
        .join(res_2, on="lit")
        .where(col("C_ACCTBAL") > col("avg_acctbal"))
        .groupby("cntrycode")
        .agg(
            col("C_ACCTBAL").count().alias("numcust"),
            col("C_ACCTBAL").sum().alias("totacctbal"),
        )
        .sort("cntrycode")
    )

    return daft_df



from __future__ import annotations

import daft
from daft import col


def test_utf8_exprs():
    df = daft.from_pydict(
        {
            "a": [
                "a",
                "df_daft",
                "foo",
                "bar",
                "baz",
                "lorém",
                "ipsum",
                "dolor",
                "sit",
                "amet",
                "😊",
                "🌟",
                "🎉",
                "This is a longer with some words",
                "THIS is ",
                "",
            ],
        }
    )

    sql = """
    SELECT
        ends_with(a, 'a') as ends_with_a,
        starts_with(a, 'a') as starts_with_a,
        contains(a, 'a') as contains_a,
        split(a, ' ') as split_a,
        regexp_match(a, 'ba.') as match_a,
        regexp_extract(a, 'ba.') as extract_a,
        regexp_extract_all(a, 'ba.') as extract_all_a,
        regexp_replace(a, 'ba.', 'foo') as replace_a,
        regexp_split(a, '\\s+') as regexp_split_a,
        length(a) as length_a,
        length_bytes(a) as length_bytes_a,
        lower(a) as lower_a,
        lstrip(a) as lstrip_a,
        rstrip(a) as rstrip_a,
        reverse(a) as reverse_a,
        capitalize(a) as capitalize_a,
        left(a, 4) as left_a,
        right(a, 4) as right_a,
        find(a, 'a') as find_a,
        rpad(a, 10, '<') as rpad_a,
        lpad(a, 10, '>') as lpad_a,
        repeat(a, 2) as repeat_a,
        a like 'a%' as like_a,
        a ilike 'a%' as ilike_a,
        substring(a, 2, 3) as substring_a,
        count_matches(a, 'a') as count_matches_a_0,
        count_matches(a, 'a', case_sensitive := true) as count_matches_a_1,
        count_matches(a, 'a', case_sensitive := false, whole_words := false) as count_matches_a_2,
        count_matches(a, 'a', case_sensitive := true, whole_words := true) as count_matches_a_3,
        normalize(a) as normalize_a,
        normalize(a, remove_punct:=true) as normalize_remove_punct_a,
        normalize(a, remove_punct:=true, lowercase:=true) as normalize_remove_punct_lower_a,
        normalize(a, remove_punct:=true, lowercase:=true, white_space:=true) as normalize_remove_punct_lower_ws_a,
        tokenize_encode(a, 'r50k_base') as tokenize_encode_a,
        tokenize_decode(tokenize_encode(a, 'r50k_base'), 'r50k_base') as tokenize_decode_a,
        concat(a, '---') as concat_a,
        concat('--', a, a, a, '--') as concat_multi_a
    FROM df
    """
    actual = daft.sql(sql).collect()
    expected = (
        df.select(
            col("a").str.endswith("a").alias("ends_with_a"),
            col("a").str.startswith("a").alias("starts_with_a"),
            col("a").str.contains("a").alias("contains_a"),
            col("a").str.split(" ").alias("split_a"),
            col("a").str.match("ba.").alias("match_a"),
            col("a").str.extract("ba.").alias("extract_a"),
            col("a").str.extract_all("ba.").alias("extract_all_a"),
            col("a").str.split(r"\s+", regex=True).alias("regexp_split_a"),
            col("a").str.replace("ba.", "foo", regex=True).alias("replace_a"),
            col("a").str.length().alias("length_a"),
            col("a").str.length_bytes().alias("length_bytes_a"),
            col("a").str.lower().alias("lower_a"),
            col("a").str.lstrip().alias("lstrip_a"),
            col("a").str.rstrip().alias("rstrip_a"),
            col("a").str.reverse().alias("reverse_a"),
            col("a").str.capitalize().alias("capitalize_a"),
            col("a").str.left(4).alias("left_a"),
            col("a").str.right(4).alias("right_a"),
            col("a").str.find("a").alias("find_a"),
            col("a").str.rpad(10, "<").alias("rpad_a"),
            col("a").str.lpad(10, ">").alias("lpad_a"),
            col("a").str.repeat(2).alias("repeat_a"),
            col("a").str.like("a%").alias("like_a"),
            col("a").str.ilike("a%").alias("ilike_a"),
            col("a").str.substr(1, 3).alias("substring_a"),
            col("a").str.count_matches("a").alias("count_matches_a_0"),
            col("a").str.count_matches("a", case_sensitive=True).alias("count_matches_a_1"),
            col("a").str.count_matches("a", case_sensitive=False, whole_words=False).alias("count_matches_a_2"),
            col("a").str.count_matches("a", case_sensitive=True, whole_words=True).alias("count_matches_a_3"),
            col("a").str.normalize().alias("normalize_a"),
            col("a").str.normalize(remove_punct=True).alias("normalize_remove_punct_a"),
            col("a").str.normalize(remove_punct=True, lowercase=True).alias("normalize_remove_punct_lower_a"),
            col("a")
            .str.normalize(remove_punct=True, lowercase=True, white_space=True)
            .alias("normalize_remove_punct_lower_ws_a"),
            col("a").str.tokenize_encode("r50k_base").alias("tokenize_encode_a"),
            col("a").str.tokenize_encode("r50k_base").str.tokenize_decode("r50k_base").alias("tokenize_decode_a"),
            col("a").str.concat("---").alias("concat_a"),
            daft.lit("--")
            .str.concat(col("a"))
            .str.concat(col("a"))
            .str.concat(col("a"))
            .str.concat("--")
            .alias("concat_multi_a"),
        )
        .collect()
        .to_pydict()
    )
    actual = actual.to_pydict()
    assert actual == expected



from __future__ import annotations

import pytest

import daft


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
def test_map_groups(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": [1, 1, 2],
            "a": [1, 3, 3],
            "b": [5, 6, 7],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.list(daft.DataType.float64()))
    def udf(a, b):
        a, b = a.to_pylist(), b.to_pylist()
        res = []
        for i in range(len(a)):
            res.append(a[i] / sum(a) + b[i])
        res.sort()
        return [res]

    daft_df = daft_df.groupby("group").map_groups(udf(daft_df["a"], daft_df["b"])).sort("group", desc=False)
    expected = {
        "group": [1, 2],
        "a": [[5.25, 6.75], [8.0]],
    }

    daft_cols = daft_df.to_pydict()

    assert daft_cols == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 3])
@pytest.mark.parametrize("output_when_empty", [[], [1], [1, 2]])
def test_map_groups_more_than_one_output_row(make_df, repartition_nparts, output_when_empty, with_morsel_size):
    daft_df = make_df(
        {
            "group": [1, 2],
            "a": [1, 3],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.int64())
    def udf(a):
        a = a.to_pylist()
        if len(a) == 0:
            return output_when_empty
        return [a[0]] * 3

    daft_df = daft_df.groupby("group").map_groups(udf(daft_df["a"])).sort("group", desc=False)
    expected = {"group": [1, 1, 1, 2, 2, 2], "a": [1, 1, 1, 3, 3, 3]}

    daft_cols = daft_df.to_pydict()

    assert daft_cols == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
def test_map_groups_single_group(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": [1, 1, 1],
            "a": [1, 2, 3],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.int64())
    def udf(a):
        a = a.to_pylist()
        if len(a) == 0:
            return []
        return [sum(x**2 for x in a)]

    daft_df = daft_df.groupby("group").map_groups(udf(daft_df["a"]))
    expected = {"group": [1], "a": [14]}

    daft_cols = daft_df.to_pydict()

    assert daft_cols == expected


@pytest.mark.parametrize("repartition_nparts", [1, 5, 11])
def test_map_groups_double_group_by(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group_1": [1, 1, 1, 1, 1, 2, 2, 2, 2, 2],
            "group_2": [1, 1, 1, 2, 2, 1, 1, 1, 2, 2],
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.int64())
    def udf(a):
        a = a.to_pylist()
        if len(a) == 0:
            return []
        return [sum(x**2 for x in a)]

    daft_df = daft_df.groupby("group_1", "group_2").map_groups(udf(daft_df["a"]))
    expected = {"group_1": [2, 2, 1, 1], "group_2": [1, 2, 1, 2], "a": [149, 181, 14, 41]}

    rows = set()
    for i in range(4):
        rows.add((expected["group_1"][i], expected["group_2"][i], expected["a"][i]))

    daft_cols = daft_df.to_pydict()
    for i in range(4):
        assert (daft_cols["group_1"][i], daft_cols["group_2"][i], daft_cols["a"][i]) in rows


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_map_groups_compound_input(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": [1, 1, 2, 2],
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.int64())
    def udf(data):
        data = data.to_pylist()
        if len(data) == 0:
            return []
        return [sum(x**2 for x in data)]

    daft_df = daft_df.groupby("group").map_groups(udf(daft_df["a"].alias("c") * daft_df["b"])).sort("group", desc=True)
    expected = {"group": [2, 1], "c": [1465, 169]}

    daft_cols = daft_df.to_pydict()
    assert daft_cols == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
def test_map_groups_with_alias(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": [1, 1, 2],
            "a": [1, 3, 3],
            "b": [5, 6, 7],
        },
        repartition=repartition_nparts,
    )

    @daft.udf(return_dtype=daft.DataType.list(daft.DataType.float64()))
    def udf(a, b):
        a, b = a.to_pylist(), b.to_pylist()
        res = []
        for i in range(len(a)):
            res.append(a[i] / sum(a) + b[i])
        res.sort()
        return [res]

    daft_df = (
        daft_df.groupby(daft_df["group"].alias("group_alias"))
        .map_groups(udf(daft_df["a"], daft_df["b"]))
        .sort("group_alias", desc=False)
    )
    expected = {
        "group_alias": [1, 2],
        "a": [[5.25, 6.75], [8.0]],
    }

    daft_cols = daft_df.to_pydict()

    assert daft_cols == expected



from __future__ import annotations

import pytest

import daft
from daft.catalog import Catalog, Identifier, NotFoundError, Table
from daft.session import Session

###
# SESSION SETUP
###


def test_current_session_exists():
    assert daft.current_session() is not None


###
# ATTACH & DETACH CATALOG
###


def test_attach_catalog():
    sess = Session()
    #
    # create some 'existing' catalogs
    cat1 = Catalog.from_pydict({})
    cat2 = Catalog.from_pydict({})
    #
    # attach them..
    sess.attach_catalog(cat1, alias="cat1")
    sess.attach_catalog(cat2, alias="cat2")
    #
    # list_catalogs
    assert 2 == len(sess.list_catalogs())
    #
    # get_catalog
    assert sess.get_catalog("cat1") == cat1
    assert sess.get_catalog("cat2") == cat2
    #
    # error!
    with pytest.raises(Exception, match="already exists"):
        sess.attach_catalog(cat1, alias="cat1")


def test_detach_catalog():
    sess = Session()
    #
    # setup.
    cat1 = Catalog.from_pydict({})
    cat2 = Catalog.from_pydict({})
    sess.attach_catalog(cat1, alias="cat1")
    sess.attach_catalog(cat2, alias="cat2")
    #
    #
    assert 2 == len(sess.list_catalogs())
    #
    # detach existing
    sess.detach_catalog("cat1")
    assert 1 == len(sess.list_catalogs())
    #
    # error!
    with pytest.raises(Exception, match="not found"):
        sess.detach_catalog("cat1")


###
# ATTACH & DETACH TABLE
###


def test_attach_table():
    sess = Session()
    #
    # create some tables
    view1 = Table.from_df("view1", daft.from_pydict({"x": [1.1, 2.2, 3.3]}))
    view2 = Table.from_df("view2", daft.from_pydict({"y": ["a", "b", "c"]}))
    #
    # attach them..
    sess.attach_table(view1, alias="tbl1")
    sess.attach_table(view2, alias="tbl2")
    #
    # list_tables
    assert 2 == len(sess.list_tables())
    #
    # get_table exact object (proper unwrapping)
    assert sess.get_table("tbl1") == view1
    assert sess.get_table("tbl2") == view2
    # read_table
    assert sess.read_table("tbl1").to_pydict() == view1.read().to_pydict()
    assert sess.read_table("tbl2").to_pydict() == view2.read().to_pydict()
    #
    # error!
    with pytest.raises(Exception, match="already exists"):
        sess.attach_table(view1, alias="tbl1")


def test_detach_table():
    sess = Session()
    #
    # setup.
    view1 = Table.from_df("view1", daft.from_pydict({"x": [1.1, 2.2, 3.3]}))
    sess.attach_table(view1, alias="tbl1")
    #
    # check is attached
    assert 1 == len(sess.list_tables())
    #
    # detach existing
    sess.detach_table("tbl1")
    assert 0 == len(sess.list_tables())
    #
    # error!
    with pytest.raises(Exception, match="not found"):
        sess.detach_table("tbl1")


###
# CREATE TABLE
###


def test_create_temp_table():
    sess = Session()
    #
    # create some dataframes
    df1 = daft.from_pydict({"x": [1.1, 2.2, 3.3]})
    df2 = daft.from_pydict({"y": ["a", "b", "c"]})
    #
    # create temp tables from these views
    sess.create_temp_table("tbl1", df1)
    sess.create_temp_table("tbl2", df2)
    #
    # list_tables
    assert 2 == len(sess.list_tables())
    #
    # read_table
    assert sess.read_table("tbl1").to_pydict() == df1.to_pydict()
    assert sess.read_table("tbl2").to_pydict() == df2.to_pydict()
    #
    # replace tbl1 with df2
    sess.create_temp_table("tbl1", df2)
    assert sess.read_table("tbl1").to_pydict() == df2.to_pydict()


###
# USE / SET [CATALOG|SCHEMA]
###


def test_use():
    sess = Session()
    #
    # create some catalogs to use
    cat1 = Catalog.from_pydict(name="cat1", tables={})
    cat2 = Catalog.from_pydict(name="cat2", tables={})
    sess.attach(cat1)
    sess.attach(cat2)
    #
    # current catalog defaults to the first attached
    assert sess.current_catalog() == cat1
    #
    # set and assert
    sess.set_catalog("cat2")
    assert sess.current_catalog() == cat2
    assert sess.current_namespace() is None
    #
    # set a namespace
    sess.set_namespace("a.b")
    assert sess.current_namespace() == Identifier("a", "b")
    #
    sess.set_catalog(None)
    sess.set_namespace(None)
    assert sess.current_catalog() is None
    assert sess.current_namespace() is None
    #
    # test use <catalog>
    sess.use("cat2")
    assert sess.current_catalog() == cat2
    assert sess.current_namespace() is None
    #
    #  test use <catalog>.<namespace>
    sess.use("cat2.a.b")
    assert sess.current_catalog() == cat2
    assert sess.current_namespace() == Identifier("a", "b")


def test_sql():
    sess = Session()
    #
    # create some catalogs to use
    cat1 = Catalog.from_pydict(name="cat1", tables={})
    cat2 = Catalog.from_pydict(name="cat2", tables={})
    sess.attach(cat1)
    sess.attach(cat2)
    #
    # test use <catalog>
    sess.sql("USE cat2")
    assert sess.current_catalog() == cat2
    assert sess.current_namespace() is None
    #
    #  test use <catalog>.<namespace>
    sess.sql("USE cat2.a.b")
    assert sess.current_catalog() == cat2
    assert sess.current_namespace() == Identifier("a", "b")


###
# exception testing
###


def test_exception_surfacing():
    class ThrowingCatalog(Catalog):
        @property
        def name(self):
            return "throwing"

        def _create_namespace(self, identifier):
            raise NotImplementedError

        def _create_table(self, identifier, source):
            raise NotImplementedError

        def _drop_namespace(self, identifier):
            raise NotImplementedError

        def _drop_table(self, identifier):
            raise NotImplementedError

        def _get_table(self, identifier):
            if str(identifier) == "boom":
                raise RuntimeError("something went wrong")
            raise NotFoundError(f"Table {identifier} not found")

        def _list_namespaces(self, pattern=None):
            raise NotImplementedError

        def _list_tables(self, pattern=None):
            raise NotImplementedError

        def _has_namespace(self, ident):
            raise NotImplementedError

        def _has_table(self, ident):
            if str(ident) == "boom":
                raise RuntimeError("something went wrong")
            raise NotFoundError(f"Table {ident} not found")

    sess = Session()
    sess.attach(ThrowingCatalog())

    # the session should
    with pytest.raises(Exception, match="not found"):
        sess.read_table("test")

    # some internal error should be surfaced in the runtime exception
    with pytest.raises(Exception, match="something went wrong"):
        sess.read_table("boom")



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft import DataType, col, lit
from daft.recordbatch import MicroPartition


def test_table_filter_all_pass() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    exprs = [col("a") < col("b"), col("a") < 5]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 4
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == [1, 2, 3, 4]
    assert result["b"] == [5, 6, 7, 8]

    exprs = [lit(True), lit(True)]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 4
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == [1, 2, 3, 4]
    assert result["b"] == [5, 6, 7, 8]


def test_table_filter_some_pass() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    exprs = [((col("a") * 4) < col("b")) | (col("b") == 8)]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == [1, 4]
    assert result["b"] == [5, 8]

    exprs = [(col("b") / col("a")) >= 3]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == [1, 2]
    assert result["b"] == [5, 6]


def test_table_filter_none_pass() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    exprs = [col("a") < col("b"), col("a") > 5]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 0
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == []
    assert result["b"] == []

    exprs = [col("a") < col("b"), lit(False)]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 0
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == []
    assert result["b"] == []


def test_table_filter_with_null_filter() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    exprs = [lit(None).cast(DataType.bool())]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 0
    assert new_table.column_names() == ["a", "b"]
    result = new_table.to_pydict()
    assert result["a"] == []
    assert result["b"] == []


def test_table_filter_bad_expression() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    exprs = [col("a") + 1]

    with pytest.raises(ValueError, match="Boolean Series"):
        daft_recordbatch.filter(exprs)


def test_table_filter_with_dates() -> None:
    from datetime import date

    def date_maker(d):
        if d is None:
            return None
        return date(2023, 1, d)

    days = list(map(date_maker, [5, 4, 1, None, 2, None]))
    pa_table = pa.Table.from_pydict({"days": days, "enum": [0, 1, 2, 3, 4, 5]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 6
    assert daft_recordbatch.column_names() == ["days", "enum"]

    exprs = [(col("days") > date(2023, 1, 2)) & (col("enum") > 0)]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 1
    assert new_table.column_names() == ["days", "enum"]
    result = new_table.to_pydict()
    assert result["days"] == [date(2023, 1, 4)]
    assert result["enum"] == [1]


def test_table_filter_with_date_days() -> None:
    from datetime import date

    def date_maker(d):
        if d is None:
            return None
        return date(2023, 1, d)

    days = list(map(date_maker, [3, 28, None, 9, 18, None]))
    pa_table = pa.Table.from_pydict({"days": days, "enum": [0, 1, 2, 3, 4, 5]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 6
    assert daft_recordbatch.column_names() == ["days", "enum"]

    exprs = [col("days").dt.day() > 15]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["days", "enum"]
    result = new_table.to_pydict()
    assert result["days"] == [date(2023, 1, 28), date(2023, 1, 18)]
    assert result["enum"] == [1, 4]


def test_table_filter_with_date_months() -> None:
    from datetime import date

    def date_maker(m):
        if m is None:
            return None
        return date(2023, m, 1)

    days = list(map(date_maker, [2, 6, None, 4, 11, None]))
    pa_table = pa.Table.from_pydict({"days": days, "enum": [0, 1, 2, 3, 4, 5]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 6
    assert daft_recordbatch.column_names() == ["days", "enum"]

    exprs = [col("days").dt.month() > 5]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["days", "enum"]
    result = new_table.to_pydict()
    assert result["days"] == [date(2023, 6, 1), date(2023, 11, 1)]
    assert result["enum"] == [1, 4]


def test_table_filter_with_date_years() -> None:
    from datetime import date

    def date_maker(y):
        if y is None:
            return None
        return date(y, 1, 1)

    days = list(map(date_maker, [5, 4000, 1, None, 2022, None]))
    pa_table = pa.Table.from_pydict({"days": days, "enum": [0, 1, 2, 3, 4, 5]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 6
    assert daft_recordbatch.column_names() == ["days", "enum"]

    exprs = [col("days").dt.year() > 2000]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["days", "enum"]
    result = new_table.to_pydict()
    assert result["days"] == [date(4000, 1, 1), date(2022, 1, 1)]
    assert result["enum"] == [1, 4]


def test_table_filter_with_date_days_of_week() -> None:
    from datetime import date

    def date_maker(d):
        if d is None:
            return None
        return date(2023, 4, d)

    # 04/03/2023 is a Monday.
    days = list(map(date_maker, [8, 5, None, 15, 12, None]))
    pa_table = pa.Table.from_pydict({"days": days, "enum": [0, 1, 2, 3, 4, 5]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 6
    assert daft_recordbatch.column_names() == ["days", "enum"]

    exprs = [col("days").dt.day_of_week() == 2]
    new_table = daft_recordbatch.filter(exprs)
    assert len(new_table) == 2
    assert new_table.column_names() == ["days", "enum"]
    result = new_table.to_pydict()
    assert result["days"] == [date(2023, 4, 5), date(2023, 4, 12)]
    assert result["enum"] == [1, 4]


def test_table_float_is_nan() -> None:
    table = MicroPartition.from_pydict({"a": [1.0, np.nan, 3.0, None, float("nan")]})
    result_table = table.eval_expression_list([col("a").float.is_nan()])
    # Note that null entries are _not_ treated as float NaNs.
    assert result_table.to_pydict() == {"a": [False, True, False, None, True]}


def test_table_float_is_inf() -> None:
    table = MicroPartition.from_pydict({"a": [-np.inf, 0.0, None, float("inf")]})
    result_table = table.eval_expression_list([col("a").float.is_inf()])
    # Note that null entries are _not_ treated as float NaNs.
    assert result_table.to_pydict() == {"a": [True, False, None, True]}


def test_table_float_not_nan() -> None:
    table = MicroPartition.from_pydict({"a": [1.0, np.nan, 3.0, None, float("nan")]})
    result_table = table.eval_expression_list([col("a").float.not_nan()])
    # Note that null entries are _not_ treated as float NaNs.
    assert result_table.to_pydict() == {"a": [True, False, True, None, False]}


def test_table_if_else() -> None:
    table = MicroPartition.from_arrow(
        pa.Table.from_pydict({"ones": [1, 1, 1], "zeros": [0, 0, 0], "pred": [True, False, None]})
    )
    result_table = table.eval_expression_list([col("pred").if_else(col("ones"), col("zeros"))])
    assert result_table.to_pydict() == {"ones": [1, 0, None]}



from __future__ import annotations

import os
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Callable, Union

from daft.dependencies import pa

if TYPE_CHECKING:
    from daft.expressions import Expression

# Column input type definitions
ColumnInputType = Union["Expression", str]
ManyColumnsInputType = Union[ColumnInputType, Iterable[ColumnInputType]]


def get_arrow_version() -> tuple[int, ...]:
    return tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


def in_notebook() -> bool:
    """Check if we are in a Jupyter notebook."""
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def pydict_to_rows(pydict: dict[str, list[Any]]) -> list[frozenset[tuple[str, Any]]]:
    """Converts a dataframe pydict to a list of rows representation.

    e.g.
    {
        "fruit": ["apple", "banana", "carrot"],
        "number": [1, 2, 3],
    }

    becomes
    [
        {("fruit", "apple"), ("number", 1)},
        {("fruit", "banana"), ("number", 2)},
        {("fruit", "carrot"), ("number", 3)},
    ]
    """
    return [
        frozenset((key, freeze(value)) for key, value in zip(pydict.keys(), values)) for values in zip(*pydict.values())
    ]


def freeze(
    input: dict[Any, Any] | list[Any] | Any,
) -> frozenset[Any] | tuple[Any, ...] | Any:
    """Freezes mutable containers for equality comparison."""
    if isinstance(input, dict):
        return frozenset((key, freeze(value)) for key, value in input.items())
    elif isinstance(input, list):
        return tuple(freeze(item) for item in input)
    else:
        return input


def map_operator_arrow_semantics_bool(
    operator: Callable[[Any, Any], Any],
    left_pylist: list[Any],
    right_pylist: list[Any],
) -> list[bool | None]:
    return [
        (bool(operator(left, right)) if (left is not None and right is not None) else None)
        for (left, right) in zip(left_pylist, right_pylist)
    ]


def python_list_membership_check(
    left_pylist: list[Any],
    right_pylist: list[Any],
) -> list[Any]:
    try:
        right_pyset = set(right_pylist)
        return [elem in right_pyset for elem in left_pylist]
    except TypeError:
        return [elem in right_pylist for elem in left_pylist]


def python_list_between_check(value_pylist: list[Any], lower_pylist: list[Any], upper_pylist: list[Any]) -> list[Any]:
    return [value <= upper and value >= lower for value, lower, upper in zip(value_pylist, lower_pylist, upper_pylist)]


def map_operator_arrow_semantics(
    operator: Callable[[Any, Any], Any],
    left_pylist: list[Any],
    right_pylist: list[Any],
) -> list[Any]:
    return [
        operator(left, right) if (left is not None and right is not None) else None
        for (left, right) in zip(left_pylist, right_pylist)
    ]


def pyarrow_supports_fixed_shape_tensor() -> bool:
    """Whether pyarrow supports the fixed_shape_tensor canonical extension type."""
    return hasattr(pa, "fixed_shape_tensor")


# Column utility functions
def is_column_input(x: Any) -> bool:
    from daft.expressions import Expression

    return isinstance(x, str) or isinstance(x, Expression)


def column_inputs_to_expressions(columns: ManyColumnsInputType) -> list[Expression]:
    """Inputs to dataframe operations can be passed in as individual arguments or an iterable.

    In addition, they may be strings or Expressions.
    This method normalizes the inputs to a list of Expressions.
    """
    from daft.expressions import col

    column_iter: Iterable[ColumnInputType] = [columns] if is_column_input(columns) else columns  # type: ignore
    return [col(c) if isinstance(c, str) else c for c in column_iter]


def detect_ray_state() -> bool:
    ray_is_initialized = False
    ray_is_in_job = False
    in_ray_worker = False
    try:
        import ray

        if ray.is_initialized():
            ray_is_initialized = True
            # Check if running inside a Ray worker
            if ray._private.worker.global_worker.mode == ray.WORKER_MODE:
                in_ray_worker = True
        # In a Ray job, Ray might not be initialized yet but we can pick up an environment variable as a heuristic here
        elif os.getenv("RAY_JOB_ID") is not None:
            ray_is_in_job = True

    except ImportError:
        pass

    return not in_ray_worker and (ray_is_initialized or ray_is_in_job)



from __future__ import annotations

import pytest


def test_sample_fraction(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df = df.sample(fraction=0.5)
    df.collect()

    assert len(df) == 2
    assert df.column_names == list(valid_data[0].keys())


def test_sample_negative_fraction(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    with pytest.raises(ValueError, match="fraction should be between 0.0 and 1.0"):
        df = df.sample(fraction=-0.1)


def test_sample_fraction_above_1(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    with pytest.raises(ValueError, match="fraction should be between 0.0 and 1.0"):
        df = df.sample(fraction=1.1)


def test_sample_full_fraction(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df = df.sample(fraction=1.0)
    df.collect()

    assert len(df) == len(valid_data)
    assert df.column_names == list(valid_data[0].keys())


def test_sample_empty(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df = df.sample(fraction=0.0)
    df.collect()

    assert len(df) == 0


def test_sample_very_small_fraction(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df = df.sample(fraction=0.0001)
    df.collect()

    # Sample with a very small fraction should still return at least a single row.
    assert len(df) == 1
    assert df.column_names == list(valid_data[0].keys())


def test_sample_with_seed(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df1 = df.sample(fraction=0.5, seed=42)
    df2 = df.sample(fraction=0.5, seed=42)

    df1.collect()
    df2.collect()

    assert len(df1) == len(df2)
    assert df1.column_names == df2.column_names == list(valid_data[0].keys())
    assert df1.to_pydict() == df2.to_pydict()


def test_sample_with_replacement(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)

    df = df.sample(fraction=0.5, with_replacement=True, seed=42)
    df.collect()

    assert len(df) == 2
    assert df.column_names == list(valid_data[0].keys())
    # Check that the two rows are the same, which should be for this seed.
    assert all(col[0] == col[1] for col in df.to_pydict().values())


def test_sample_without_replacement(make_df, valid_data: list[dict[str, float]]) -> None:
    # Sample without replacement should return different rows each time.
    # Valid data has 3 rows, so 10 iterations should be enough to test this.
    for _ in range(10):
        df = make_df(valid_data)
        df = df.sample(fraction=0.5, with_replacement=False)
        df.collect()

        assert len(df) == 2
        assert df.column_names == list(valid_data[0].keys())
        # Check that the two rows are different.
        pylist = df.to_pylist()
        assert pylist[0] != pylist[1]


def test_sample_with_concat(make_df, valid_data: list[dict[str, float]]) -> None:
    df1 = make_df(valid_data)
    df2 = make_df(valid_data)

    df1 = df1.sample(fraction=0.5, seed=42)
    df2 = df2.sample(fraction=0.5, seed=42)

    df = df1.concat(df2)
    df.collect()

    assert len(df) == 4
    assert df.column_names == list(valid_data[0].keys())
    # Check that the two rows are the same, which should be for this seed.
    assert all(col[:2] == col[2:] for col in df.to_pydict().values())



from __future__ import annotations

from typing import Any, Literal

import pytest

import daft
from daft import col

API = Literal["Expression", "SQL", "Series"]


@pytest.fixture(scope="function")
def series(request):
    """Returns a series-like fixture backed by the respective API's function provider for batch function testing."""

    def _series(input: list[Any]):
        return {
            "Expression": _ExpressionFunctions,
            "SQL": _SqlFunctions,
            "Series": _SeriesFunctions,
        }[request.param](input=input)

    return _series


class _ExpressionFunctions:
    """Expression API function helper."""

    def __init__(self, input: list[Any]):
        self.input = input

    def __getattribute__(self, name):
        def wrapper(*args, **kwargs):
            expr = col("input").__getattribute__(name)(*args, **kwargs)
            df = daft.from_pydict({"input": object.__getattribute__(self, "input")})
            df = df.select(expr.alias("output")).to_pydict()
            return df["output"]

        return wrapper


class _SqlFunctions:
    """SQL API function helper."""

    def __init__(self, input: list[Any]):
        self.input = input

    def __getattribute__(self, name):
        def wrapper(*args, **kwargs):
            expr = f"{name}({', '.join(to_sql_args(*args, **kwargs))})"
            df = daft.from_pydict({"input": object.__getattribute__(self, "input")})
            df = daft.sql(f"SELECT {expr} AS output FROM df", df=df).to_pydict()
            return df["output"]

        return wrapper


class _SeriesFunctions:
    """Series API function helper."""

    def __init__(self, input: list[Any]):
        self.input = input

    def __getattribute__(self, name):
        raise ValueError("not supported")


def to_sql_args(*args, **kwargs) -> list[str]:
    sql_args = ["input"]
    sql_args.extend([to_sql_arg(arg) for arg in args])
    sql_args.extend([f"{name}:={to_sql_arg(arg)}" for (name, arg) in kwargs.items()])
    return sql_args


def to_sql_arg(arg) -> str:
    if arg is None:
        return "NULL"
    elif isinstance(arg, str):
        return f"'{arg}'"
    elif isinstance(arg, list):
        items = [to_sql_arg(arg) for item in arg]
        return f"[{', '.join(items)}]"
    else:
        raise ValueError("Unsupported SQL argument!")
